@@ -60,14 +60,22 @@
 #define NETBSD_PARTITION	0xa9
 #define DKTYPENAMES
 #include "fdiskbsdlabel.h"
+#include "fdiskdoslabel.h"
 
-static void xbsd_delete_part (struct fdisk_context *cxt, int partnum);
-static void xbsd_edit_disklabel (void);
+/*
+ * in-memory fdisk BSD stuff
+ */
+struct fdisk_bsd_label {
+	struct fdisk_label	head;		/* generic part */
+};
+
+
+static int xbsd_delete_part (struct fdisk_context *cxt, size_t partnum);
+static void xbsd_edit_disklabel (struct fdisk_context *cxt);
 static void xbsd_write_bootstrap (struct fdisk_context *cxt);
-static void xbsd_change_fstype (void);
-static int xbsd_get_part_index (int max);
-static int xbsd_check_new_partition (int *i);
-static void xbsd_list_types (void);
+static void xbsd_change_fstype (struct fdisk_context *cxt);
+static int xbsd_get_part_index (struct fdisk_context *cxt, int max);
+static int xbsd_check_new_partition (struct fdisk_context *cxt, int *i);
 static unsigned short xbsd_dkcksum (struct xbsd_disklabel *lp);
 static int xbsd_initlabel  (struct fdisk_context *cxt,
 			    struct partition *p, struct xbsd_disklabel *d,
@@ -97,8 +105,8 @@ static char disklabelbuffer[BSD_BBSIZE];
 
 static struct xbsd_disklabel xbsd_dlabel;
 
-#define bsd_cround(n) \
-	(display_in_cyl_units ? ((n)/xbsd_dlabel.d_secpercyl) + 1 : (n))
+#define bsd_cround(c, n) \
+	(fdisk_context_use_cylinders(c) ? ((n)/xbsd_dlabel.d_secpercyl) + 1 : (n))
 
 /*
  * Test whether the whole disk has BSD disk label magic.
@@ -107,29 +115,25 @@ static struct xbsd_disklabel xbsd_dlabel;
  * so this does not mean that there is a BSD disk label.
  */
 static int
-osf_probe_label(struct fdisk_context *cxt) {
+osf_probe_label(struct fdisk_context *cxt)
+{
+	assert(cxt);
+	assert(cxt->label);
+	assert(fdisk_is_disklabel(cxt, OSF));
+
 	if (xbsd_readlabel (cxt, NULL, &xbsd_dlabel) == 0)
 		return 0;
 	return 1;
 }
 
-int
-btrydev (struct fdisk_context *cxt) {
-	if (xbsd_readlabel (cxt, NULL, &xbsd_dlabel) == 0)
-		return -1;
-	printf(_("\nBSD label for device: %s\n"), cxt->dev_path);
-	xbsd_print_disklabel (cxt, 0);
-	return 0;
-}
-
 #if !defined (__alpha__)
-static int
-hidden(int type) {
+static int hidden(int type)
+{
 	return type ^ 0x10;
 }
 
-static int
-is_bsd_partition_type(int type) {
+static int is_bsd_partition_type(int type)
+{
 	return (type == FREEBSD_PARTITION ||
 		type == hidden(FREEBSD_PARTITION) ||
 		type == NETBSD_PARTITION ||
@@ -139,12 +143,10 @@ is_bsd_partition_type(int type) {
 
 static int xbsd_write_disklabel (struct fdisk_context *cxt)
 {
-#if defined (__alpha__)
 	printf (_("Writing disklabel to %s.\n"), cxt->dev_path);
+#if defined (__alpha__)
 	xbsd_writelabel (cxt, NULL, &xbsd_dlabel);
 #else
-	printf (_("Writing disklabel to %s.\n"),
-		partname(cxt->dev_path, xbsd_part_index+1, 0));
 	xbsd_writelabel (cxt, xbsd_part, &xbsd_dlabel);
 #endif
 	reread_partition_table(cxt, 0);	/* no exit yet */
@@ -152,16 +154,21 @@ static int xbsd_write_disklabel (struct fdisk_context *cxt)
 	return 0;
 }
 
-static void xbsd_add_part (struct fdisk_context *cxt,
-		int partnum __attribute__((__unused__)),
-		int parttype __attribute__((__unused__)))
+static int xbsd_add_part (struct fdisk_context *cxt,
+		size_t partnum __attribute__((__unused__)),
+		struct fdisk_parttype *t __attribute__((__unused__)))
 {
 	unsigned int begin, end;
 	char mesg[256];
-	int i;
+	int i, rc;
 
-	if (!xbsd_check_new_partition (&i))
-		return;
+	assert(cxt);
+	assert(cxt->label);
+	assert(fdisk_is_disklabel(cxt, OSF));
+
+	rc = xbsd_check_new_partition(cxt, &i);
+	if (rc)
+		return rc;
 
 #if !defined (__alpha__) && !defined (__powerpc__) && !defined (__hppa__)
 	begin = get_start_sect(xbsd_part);
@@ -171,39 +178,48 @@ static void xbsd_add_part (struct fdisk_context *cxt,
 	end = xbsd_dlabel.d_secperunit - 1;
 #endif
 
-	snprintf (mesg, sizeof(mesg), _("First %s"), str_units(SINGULAR));
-	begin = read_int (cxt, bsd_cround (begin), bsd_cround (begin), bsd_cround (end),
+	snprintf (mesg, sizeof(mesg), _("First %s"),
+			fdisk_context_get_unit(cxt, SINGULAR));
+	begin = read_int(cxt, bsd_cround (cxt, begin),
+			      bsd_cround (cxt, begin),
+			      bsd_cround (cxt, end),
 			  0, mesg);
 
-	if (display_in_cyl_units)
+	if (fdisk_context_use_cylinders(cxt))
 		begin = (begin - 1) * xbsd_dlabel.d_secpercyl;
 
 	snprintf (mesg, sizeof(mesg), _("Last %s or +size or +sizeM or +sizeK"),
-		  str_units(SINGULAR));
-	end = read_int (cxt, bsd_cround (begin), bsd_cround (end), bsd_cround (end),
-			bsd_cround (begin), mesg);
+		  fdisk_context_get_unit(cxt, SINGULAR));
+	end = read_int (cxt, bsd_cround (cxt, begin),
+			bsd_cround (cxt, end),
+			bsd_cround (cxt, end),
+			bsd_cround (cxt, begin), mesg);
 
-	if (display_in_cyl_units)
+	if (fdisk_context_use_cylinders(cxt))
 		end = end * xbsd_dlabel.d_secpercyl - 1;
 
 	xbsd_dlabel.d_partitions[i].p_size   = end - begin + 1;
 	xbsd_dlabel.d_partitions[i].p_offset = begin;
 	xbsd_dlabel.d_partitions[i].p_fstype = BSD_FS_UNUSED;
+
+	cxt->label->nparts_cur = xbsd_dlabel.d_npartitions;
+	fdisk_label_set_changed(cxt->label, 1);
+
+	return 0;
 }
 
-static int xbsd_create_disklabel (struct fdisk_context *cxt)
+static int xbsd_create_disklabel(struct fdisk_context *cxt)
 {
 	char c;
 
-#if defined (__alpha__)
+	assert(cxt);
+	assert(cxt->label);
+	assert(fdisk_is_disklabel(cxt, OSF));
+
 	fprintf (stderr, _("%s contains no disklabel.\n"), cxt->dev_path);
-#else
-	fprintf (stderr, _("%s contains no disklabel.\n"),
-		 partname(cxt->dev_path, xbsd_part_index+1, 0));
-#endif
 
 	while (1) {
-		c = read_char (_("Do you want to create a disklabel? (y/n) "));
+		c = read_char(cxt, _("Do you want to create a disklabel? (y/n) "));
 		if (tolower(c) == 'y') {
 			if (xbsd_initlabel (cxt,
 #if defined (__alpha__) || defined (__powerpc__) || defined (__hppa__) || \
@@ -214,6 +230,8 @@ static int xbsd_create_disklabel (struct fdisk_context *cxt)
 #endif
 				) == 1) {
 				xbsd_print_disklabel (cxt, 1);
+				cxt->label->nparts_cur = xbsd_dlabel.d_npartitions;
+				cxt->label->nparts_max = BSD_MAXPARTITIONS;
 				return 1;
 			} else
 				return 0;
@@ -229,19 +247,28 @@ bsd_command_prompt (struct fdisk_context *cxt)
   int t, ss;
   struct partition *p;
 
+  assert(cxt);
+  assert(cxt->parent);
+
   for (t=0; t<4; t++) {
     p = get_part_table(t);
     if (p && is_bsd_partition_type(p->sys_ind)) {
       xbsd_part = p;
       xbsd_part_index = t;
       ss = get_start_sect(xbsd_part);
+
+      /* TODO - partname uses static buffer!!! */
+      cxt->dev_path = partname(cxt->parent->dev_path, t+1, 0);
+      if (cxt->dev_path)
+	      cxt->dev_path = strdup(cxt->dev_path);
+
       if (ss == 0) {
 	fprintf (stderr, _("Partition %s has invalid starting sector 0.\n"),
-		 partname(cxt->dev_path, t+1, 0));
+		 cxt->dev_path);
 	return;
       }
       printf (_("Reading disklabel of %s at sector %d.\n"),
-	      partname(cxt->dev_path, t+1, 0), ss + BSD_LABELSECTOR);
+		   cxt->dev_path, ss + BSD_LABELSECTOR);
       if (xbsd_readlabel (cxt, xbsd_part, &xbsd_dlabel) == 0)
 	if (xbsd_create_disklabel (cxt) == 0)
 	  return;
@@ -264,18 +291,18 @@ bsd_command_prompt (struct fdisk_context *cxt)
 
   while (1) {
     putchar ('\n');
-    switch (tolower (read_char (_("BSD disklabel command (m for help): ")))) {
+    switch (tolower (read_char(cxt, _("BSD disklabel command (m for help): ")))) {
       case 'd':
-	      xbsd_delete_part(cxt, xbsd_get_part_index(xbsd_dlabel.d_npartitions));
+	      xbsd_delete_part(cxt, xbsd_get_part_index(cxt, xbsd_dlabel.d_npartitions));
 	      break;
       case 'e':
-	xbsd_edit_disklabel ();
+	xbsd_edit_disklabel (cxt);
 	break;
       case 'i':
 	xbsd_write_bootstrap (cxt);
 	break;
       case 'l':
-	xbsd_list_types ();
+	list_partition_types (cxt);
 	break;
       case 'n':
 	      xbsd_add_part (cxt, 0, 0);
@@ -292,10 +319,10 @@ bsd_command_prompt (struct fdisk_context *cxt)
 	      xbsd_print_disklabel (cxt, 1);
 	break;
       case 't':
-	xbsd_change_fstype ();
+	xbsd_change_fstype (cxt);
 	break;
       case 'u':
-	change_units(cxt);
+	toggle_units(cxt);
 	break;
       case 'w':
 	xbsd_write_disklabel (cxt);
@@ -306,36 +333,42 @@ bsd_command_prompt (struct fdisk_context *cxt)
 	break;
 #endif
       default:
-	print_menu(MAIN_MENU);
+	print_menu(cxt, MAIN_MENU);
 	break;
     }
   }
 }
 
-static void xbsd_delete_part(struct fdisk_context *cxt __attribute__((__unused__)),
-			     int partnum)
+static int xbsd_delete_part(
+		struct fdisk_context *cxt,
+		size_t partnum)
 {
+	assert(cxt);
+	assert(cxt->label);
+	assert(fdisk_is_disklabel(cxt, OSF));
+
 	xbsd_dlabel.d_partitions[partnum].p_size   = 0;
 	xbsd_dlabel.d_partitions[partnum].p_offset = 0;
 	xbsd_dlabel.d_partitions[partnum].p_fstype = BSD_FS_UNUSED;
 	if (xbsd_dlabel.d_npartitions == partnum + 1)
 		while (!xbsd_dlabel.d_partitions[xbsd_dlabel.d_npartitions-1].p_size)
 			xbsd_dlabel.d_npartitions--;
+
+	cxt->label->nparts_cur = xbsd_dlabel.d_npartitions;
+	fdisk_label_set_changed(cxt->label, 1);
+	return 0;
 }
 
 void
-xbsd_print_disklabel (struct fdisk_context *cxt, int show_all) {
+xbsd_print_disklabel (struct fdisk_context *cxt, int show_all)
+{
   struct xbsd_disklabel *lp = &xbsd_dlabel;
   struct xbsd_partition *pp;
   FILE *f = stdout;
   int i, j;
 
   if (show_all) {
-#if defined (__alpha__)
     fprintf(f, "# %s:\n", cxt->dev_path);
-#else
-    fprintf(f, "# %s:\n", partname(cxt->dev_path, xbsd_part_index+1, 0));
-#endif
     if ((unsigned) lp->d_type < BSD_DKMAXTYPES)
       fprintf(f, _("type: %s\n"), xbsd_dktypenames[lp->d_type]);
     else
@@ -379,7 +412,7 @@ xbsd_print_disklabel (struct fdisk_context *cxt, int show_all) {
   pp = lp->d_partitions;
   for (i = 0; i < lp->d_npartitions; i++, pp++) {
     if (pp->p_size) {
-      if (display_in_cyl_units && lp->d_secpercyl) {
+      if (fdisk_context_use_cylinders(cxt) && lp->d_secpercyl) {
 	fprintf(f, "  %c: %8ld%c %8ld%c %8ld%c  ",
 		'a' + i,
 		(long) pp->p_offset / lp->d_secpercyl + 1,
@@ -421,49 +454,49 @@ xbsd_print_disklabel (struct fdisk_context *cxt, int show_all) {
   }
 }
 
-static int
-edit_int (int def, char *mesg)
+static unsigned long
+edit_int(struct fdisk_context *cxt, unsigned long def, char *mesg)
 {
   do {
     fputs (mesg, stdout);
-    printf (" (%d): ", def);
-    if (!read_line (NULL))
+    printf (" (%lu): ", def);
+    if (!read_line(cxt, NULL))
       return def;
-  }
-  while (!isdigit (*line_ptr));
-  return atoi (line_ptr);
+  } while (!isdigit (*line_ptr));
+
+  return strtoul(line_ptr, NULL, 10);	/* TODO check it! */
 }
 
 static void
-xbsd_edit_disklabel (void)
+xbsd_edit_disklabel(struct fdisk_context *cxt)
 {
   struct xbsd_disklabel *d;
 
   d = &xbsd_dlabel;
 
 #if defined (__alpha__) || defined (__ia64__)
-  d -> d_secsize    = (unsigned long) edit_int ((unsigned long) d -> d_secsize     ,_("bytes/sector"));
-  d -> d_nsectors   = (unsigned long) edit_int ((unsigned long) d -> d_nsectors    ,_("sectors/track"));
-  d -> d_ntracks    = (unsigned long) edit_int ((unsigned long) d -> d_ntracks     ,_("tracks/cylinder"));
-  d -> d_ncylinders = (unsigned long) edit_int ((unsigned long) d -> d_ncylinders  ,_("cylinders"));
+  d -> d_secsize    = edit_int(cxt, d->d_secsize     ,_("bytes/sector"));
+  d -> d_nsectors   = edit_int(cxt, d->d_nsectors    ,_("sectors/track"));
+  d -> d_ntracks    = edit_int(cxt, d->d_ntracks     ,_("tracks/cylinder"));
+  d -> d_ncylinders = edit_int(cxt, d->d_ncylinders  ,_("cylinders"));
 #endif
 
   /* d -> d_secpercyl can be != d -> d_nsectors * d -> d_ntracks */
   while (1)
   {
-    d -> d_secpercyl = (unsigned long) edit_int ((unsigned long) d -> d_nsectors * d -> d_ntracks,
+    d -> d_secpercyl = edit_int(cxt, (unsigned long) d->d_nsectors * d -> d_ntracks,
 					  _("sectors/cylinder"));
     if (d -> d_secpercyl <= d -> d_nsectors * d -> d_ntracks)
       break;
 
     printf (_("Must be <= sectors/track * tracks/cylinder (default).\n"));
   }
-  d -> d_rpm        = (unsigned short) edit_int ((unsigned short) d -> d_rpm       ,_("rpm"));
-  d -> d_interleave = (unsigned short) edit_int ((unsigned short) d -> d_interleave,_("interleave"));
-  d -> d_trackskew  = (unsigned short) edit_int ((unsigned short) d -> d_trackskew ,_("trackskew"));
-  d -> d_cylskew    = (unsigned short) edit_int ((unsigned short) d -> d_cylskew   ,_("cylinderskew"));
-  d -> d_headswitch = (unsigned long) edit_int ((unsigned long) d -> d_headswitch  ,_("headswitch"));
-  d -> d_trkseek    = (unsigned long) edit_int ((unsigned long) d -> d_trkseek     ,_("track-to-track seek"));
+  d -> d_rpm        = (unsigned short) edit_int(cxt, d->d_rpm       ,_("rpm"));
+  d -> d_interleave = (unsigned short) edit_int(cxt, d->d_interleave,_("interleave"));
+  d -> d_trackskew  = (unsigned short) edit_int(cxt, d->d_trackskew ,_("trackskew"));
+  d -> d_cylskew    = (unsigned short) edit_int(cxt, d->d_cylskew   ,_("cylinderskew"));
+  d -> d_headswitch = edit_int(cxt, d->d_headswitch  ,_("headswitch"));
+  d -> d_trkseek    = edit_int(cxt, d->d_trkseek     ,_("track-to-track seek"));
 
   d -> d_secperunit = d -> d_secpercyl * d -> d_ncylinders;
 }
@@ -506,7 +539,7 @@ xbsd_write_bootstrap (struct fdisk_context *cxt)
 
   printf (_("Bootstrap: %sboot -> boot%s (%s): "),
 	  dkbasename, dkbasename, dkbasename);
-  if (read_line (NULL)) {
+  if (read_line(cxt, NULL)) {
     line_ptr[strlen (line_ptr)-1] = '\0';
     dkbasename = line_ptr;
   }
@@ -549,41 +582,47 @@ xbsd_write_bootstrap (struct fdisk_context *cxt)
   if (BSD_BBSIZE != write (cxt->dev_fd, disklabelbuffer, BSD_BBSIZE))
 	  fatal (cxt, unable_to_write);
 
-#if defined (__alpha__)
   printf (_("Bootstrap installed on %s.\n"), cxt->dev_path);
-#else
-  printf (_("Bootstrap installed on %s.\n"),
-    partname (cxt->dev_path, xbsd_part_index+1, 0));
-#endif
 
   sync_disks ();
 }
 
-static void
-xbsd_change_fstype (void)
+/* TODO: remove this, use regular change_partition_type() in fdisk.c */
+static void xbsd_change_fstype (struct fdisk_context *cxt)
 {
   int i;
+  struct fdisk_parttype *t;
 
-  i = xbsd_get_part_index (xbsd_dlabel.d_npartitions);
-  xbsd_dlabel.d_partitions[i].p_fstype = read_hex (xbsd_fstypes);
+  assert(cxt);
+  assert(cxt->label);
+  assert(fdisk_is_disklabel(cxt, OSF));
+
+  i = xbsd_get_part_index (cxt, xbsd_dlabel.d_npartitions);
+  t = read_partition_type(cxt);
+
+  if (t) {
+    xbsd_dlabel.d_partitions[i].p_fstype = t->type;
+    fdisk_free_parttype(t);
+    fdisk_label_set_changed(cxt->label, 1);
+  }
 }
 
 static int
-xbsd_get_part_index (int max)
+xbsd_get_part_index(struct fdisk_context *cxt, int max)
 {
   char prompt[256];
   char l;
 
   snprintf (prompt, sizeof(prompt), _("Partition (a-%c): "), 'a' + max - 1);
   do
-     l = tolower (read_char (prompt));
+     l = tolower(read_char(cxt, prompt));
   while (l < 'a' || l > 'a' + max - 1);
   return l - 'a';
 }
 
 static int
-xbsd_check_new_partition (int *i) {
-
+xbsd_check_new_partition(struct fdisk_context *cxt, int *i)
+{
 	/* room for more? various BSD flavours have different maxima */
 	if (xbsd_dlabel.d_npartitions == BSD_MAXPARTITIONS) {
 		int t;
@@ -595,26 +634,21 @@ xbsd_check_new_partition (int *i) {
 		if (t == BSD_MAXPARTITIONS) {
 			fprintf (stderr, _("The maximum number of partitions "
 					   "has been created\n"));
-			return 0;
+			return -EINVAL;
 		}
 	}
 
-	*i = xbsd_get_part_index (BSD_MAXPARTITIONS);
+	*i = xbsd_get_part_index(cxt, BSD_MAXPARTITIONS);
 
 	if (*i >= xbsd_dlabel.d_npartitions)
 		xbsd_dlabel.d_npartitions = (*i) + 1;
 
 	if (xbsd_dlabel.d_partitions[*i].p_size != 0) {
 		fprintf (stderr, _("This partition already exists.\n"));
-		return 0;
+		return -EINVAL;
 	}
 
-	return 1;
-}
-
-static void
-xbsd_list_types (void) {
-	list_types (xbsd_fstypes);
+	return 0;
 }
 
 static unsigned short
@@ -705,6 +739,9 @@ xbsd_readlabel (struct fdisk_context *cxt, struct partition *p, struct xbsd_disk
 {
 	int t, sector;
 
+	assert(cxt);
+	assert(cxt->label);
+
 	/* p is used only to get the starting sector */
 #if !defined (__alpha__)
 	sector = (p ? get_start_sect(p) : 0);
@@ -713,9 +750,9 @@ xbsd_readlabel (struct fdisk_context *cxt, struct partition *p, struct xbsd_disk
 #endif
 
 	if (lseek (cxt->dev_fd, (off_t) sector * SECTOR_SIZE, SEEK_SET) == -1)
-		fatal (cxt, unable_to_seek);
+		return 0;
 	if (BSD_BBSIZE != read (cxt->dev_fd, disklabelbuffer, BSD_BBSIZE))
-		fatal (cxt, unable_to_read);
+		return 0;
 
 	memmove (d,
 	         &disklabelbuffer[BSD_LABELSECTOR * SECTOR_SIZE + BSD_LABELOFFSET],
@@ -734,6 +771,9 @@ xbsd_readlabel (struct fdisk_context *cxt, struct partition *p, struct xbsd_disk
 		fprintf (stderr, _("Warning: too many partitions "
 				   "(%d, maximum is %d).\n"),
 			 d -> d_npartitions, BSD_MAXPARTITIONS);
+
+	cxt->label->nparts_cur = d->d_npartitions;
+	cxt->label->nparts_max = BSD_MAXPARTITIONS;
 	return 1;
 }
 
@@ -804,22 +844,30 @@ xbsd_translate_fstype (int linux_type)
   }
 }
 
+/*
+ * link partition from parent (DOS) to nested BSD partition table
+ */
 static void
 xbsd_link_part (struct fdisk_context *cxt)
 {
-  int k, i;
-  struct partition *p;
+	size_t k;
+	int i;
+	struct partition *p;
 
-  k = get_partition (cxt, 1, partitions);
+	if (!cxt->parent)
+		return;		/* not nested PT */
 
-  if (!xbsd_check_new_partition (&i))
-    return;
+	if (fdisk_ask_partnum(cxt->parent, &k, FALSE))
+		return;
 
-  p = get_part_table(k);
+	if (xbsd_check_new_partition(cxt, &i))
+		return;
 
-  xbsd_dlabel.d_partitions[i].p_size   = get_nr_sects(p);
-  xbsd_dlabel.d_partitions[i].p_offset = get_start_sect(p);
-  xbsd_dlabel.d_partitions[i].p_fstype = xbsd_translate_fstype(p->sys_ind);
+	p = get_part_table(k);
+
+	xbsd_dlabel.d_partitions[i].p_size   = get_nr_sects(p);
+	xbsd_dlabel.d_partitions[i].p_offset = get_start_sect(p);
+	xbsd_dlabel.d_partitions[i].p_fstype = xbsd_translate_fstype(p->sys_ind);
 }
 #endif
 
@@ -843,13 +891,84 @@ alpha_bootblock_checksum (char *boot)
 }
 #endif /* __alpha__ */
 
-const struct fdisk_label bsd_label =
+static struct fdisk_parttype *xbsd_get_parttype(
+		struct fdisk_context *cxt,
+		size_t n)
 {
-	.name = "bsd",
-	.probe = osf_probe_label,
-	.write = xbsd_write_disklabel,
-	.verify = NULL,
-	.create = xbsd_create_disklabel,
-	.part_add = xbsd_add_part,
-	.part_delete = xbsd_delete_part,
+	struct fdisk_parttype *t;
+
+	assert(cxt);
+	assert(cxt->label);
+	assert(fdisk_is_disklabel(cxt, OSF));
+
+	if (n >= xbsd_dlabel.d_npartitions)
+		return NULL;
+
+	t = fdisk_get_parttype_from_code(cxt, xbsd_dlabel.d_partitions[n].p_fstype);
+	if (!t)
+		t = fdisk_new_unknown_parttype(xbsd_dlabel.d_partitions[n].p_fstype, NULL);
+	return t;
+}
+
+static int xbsd_set_parttype(
+		struct fdisk_context *cxt,
+		size_t partnum,
+		struct fdisk_parttype *t)
+{
+	struct xbsd_partition *p;
+
+	assert(cxt);
+	assert(cxt->label);
+	assert(fdisk_is_disklabel(cxt, OSF));
+
+	if (partnum >= xbsd_dlabel.d_npartitions || !t || t->type > UINT8_MAX)
+		return -EINVAL;
+
+	p = &xbsd_dlabel.d_partitions[partnum];
+	if (t->type == p->p_fstype)
+		return 0;
+
+	p->p_fstype = t->type;
+	fdisk_label_set_changed(cxt->label, 1);
+	return 0;
+}
+
+static const struct fdisk_label_operations bsd_operations =
+{
+	.probe		= osf_probe_label,
+	.write		= xbsd_write_disklabel,
+	.create		= xbsd_create_disklabel,
+	.part_add	= xbsd_add_part,
+	.part_delete	= xbsd_delete_part,
+	.part_get_type	= xbsd_get_parttype,
+	.part_set_type	= xbsd_set_parttype
 };
+
+
+/*
+ * allocates BSD label driver
+ */
+struct fdisk_label *fdisk_new_bsd_label(struct fdisk_context *cxt)
+{
+	struct fdisk_label *lb;
+	struct fdisk_bsd_label *bsd;
+
+	assert(cxt);
+
+	bsd = calloc(1, sizeof(*bsd));
+	if (!bsd)
+		return NULL;
+
+	/* initialize generic part of the driver */
+	lb = (struct fdisk_label *) bsd;
+	lb->name = "bsd";
+	lb->id = FDISK_DISKLABEL_OSF;
+	lb->op = &bsd_operations;
+	lb->parttypes = xbsd_fstypes;
+	lb->nparttypes = ARRAY_SIZE(xbsd_fstypes);
+
+	/* don't ask for partition number for op->part_add() */
+	lb->flags = FDISK_LABEL_FL_ADDPART_NOPARTNO;
+
+	return lb;
+}
