@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2010 Karel Zak <kzak@redhat.com>
+ * Copyright (C) 2010,2011,2012 Karel Zak <kzak@redhat.com>
  *
  * This file may be redistributed under the terms of the
  * GNU Lesser General Public License.
@@ -33,6 +33,8 @@
 
 #include "mountP.h"
 
+#include <sys/wait.h>
+
 /**
  * mnt_new_context:
  *
@@ -47,11 +49,13 @@ struct libmnt_context *mnt_new_context(void)
 	if (!cxt)
 		return NULL;
 
+	INIT_LIST_HEAD(&cxt->addmounts);
+
 	ruid = getuid();
 	euid = geteuid();
 
-	cxt->syscall_status = 1;	/* not called yet */
-	cxt->helper_exec_status = 1;
+	mnt_context_reset_status(cxt);
+
 	cxt->loopdev_fd = -1;
 
 	/* if we're really root and aren't running setuid */
@@ -94,6 +98,8 @@ void mnt_free_context(struct libmnt_context *cxt)
 	mnt_free_lock(cxt->lock);
 	mnt_free_update(cxt->update);
 
+	free(cxt->children);
+
 	DBG(CXT, mnt_debug_h(cxt, "<---- free"));
 	free(cxt);
 }
@@ -140,16 +146,23 @@ int mnt_reset_context(struct libmnt_context *cxt)
 
 	cxt->fs = NULL;
 	cxt->mtab = NULL;
-	cxt->ambi = 0;
 	cxt->helper = NULL;
 	cxt->orig_user = NULL;
 	cxt->mountflags = 0;
 	cxt->user_mountflags = 0;
 	cxt->mountdata = NULL;
 	cxt->flags = MNT_FL_DEFAULT;
-	cxt->syscall_status = 1;
-	cxt->helper_exec_status = 1;
-	cxt->helper_status = 0;
+
+	/* free additional mounts list */
+	while (!list_empty(&cxt->addmounts)) {
+		struct libmnt_addmount *ad = list_entry(cxt->addmounts.next,
+				                  struct libmnt_addmount,
+						  mounts);
+		mnt_free_addmount(ad);
+	}
+
+	mnt_context_reset_status(cxt);
+	mnt_context_set_tabfilter(cxt, NULL, NULL);
 
 	/* restore non-resetable flags */
 	cxt->flags |= (fl & MNT_FL_EXTERN_FSTAB);
@@ -161,20 +174,50 @@ int mnt_reset_context(struct libmnt_context *cxt)
 	cxt->flags |= (fl & MNT_FL_NOHELPERS);
 	cxt->flags |= (fl & MNT_FL_LOOPDEL);
 	cxt->flags |= (fl & MNT_FL_LAZY);
+	cxt->flags |= (fl & MNT_FL_FORK);
 	cxt->flags |= (fl & MNT_FL_FORCE);
 	cxt->flags |= (fl & MNT_FL_NOCANONICALIZE);
 	cxt->flags |= (fl & MNT_FL_RDONLY_UMOUNT);
+	cxt->flags |= (fl & MNT_FL_NOSWAPMATCH);
+	return 0;
+}
+
+/**
+ * mnt_context_reset_status:
+ * @cxt: context
+ *
+ * Resets mount(2) and mount.type statuses, so mnt_context_do_mount() or
+ * mnt_context_do_umount() could be again called with the same settings.
+ *
+ * BE CAREFUL -- after this soft reset the libmount will NOT parse mount
+ * options, evaluate permissions or apply stuff from fstab.
+ *
+ * Returns: 0 on success, negative number in case of error.
+ */
+int mnt_context_reset_status(struct libmnt_context *cxt)
+{
+	assert(cxt);
+	if (!cxt)
+		return -EINVAL;
+
+	cxt->syscall_status = 1;		/* means not called yet */
+	cxt->helper_exec_status = 1;
+	cxt->helper_status = 0;
 	return 0;
 }
 
 static int set_flag(struct libmnt_context *cxt, int flag, int enable)
 {
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
-	if (enable)
+	if (enable) {
+		DBG(CXT, mnt_debug_h(cxt, "enabling flag %04x", flag));
 		cxt->flags |= flag;
-	else
+	} else {
+		DBG(CXT, mnt_debug_h(cxt, "disabling flag %04x", flag));
 		cxt->flags &= ~flag;
+	}
 	return 0;
 }
 
@@ -186,21 +229,49 @@ static int set_flag(struct libmnt_context *cxt, int flag, int enable)
  */
 int mnt_context_is_restricted(struct libmnt_context *cxt)
 {
-	assert(cxt);
 	return cxt->restricted;
 }
 
 /**
  * mnt_context_set_optsmode
  * @cxt: mount context
- * @mode: mask, see MNT_OMASK_* flags in libmount mount.h
+ * @mode: MNT_OMASK_* flags
  *
- * Controls how to use mount options from fstab/mtab.
+ * Controls how to use mount optionsmsource and target paths from fstab/mtab.
+ *
+ * @MNT_OMODE_IGNORE: ignore mtab/fstab options
+ *
+ * @MNT_OMODE_APPEND: append mtab/fstab options to existing options
+ *
+ * @MNT_OMODE_PREPEND: prepend mtab/fstab options to existing options
+ *
+ * @MNT_OMODE_REPLACE: replace existing options with options from mtab/fstab
+ *
+ * @MNT_OMODE_FORCE: always read mtab/fstab (although source and target is defined)
+ *
+ * @MNT_OMODE_FSTAB: read from fstab
+ *
+ * @MNT_OMODE_MTAB: read from mtab if fstab not enabled or failed
+ *
+ * @MNT_OMODE_NOTAB: do not read fstab/mtab at all
+ *
+ * @MNT_OMODE_AUTO: default mode (MNT_OMODE_PREPEND | MNT_OMODE_FSTAB | MNT_OMODE_MTAB)
+ *
+ * @MNT_OMODE_USER: default for non-root users (MNT_OMODE_REPLACE | MNT_OMODE_FORCE | MNT_OMODE_FSTAB)
+ *
+ * Notes:
+ *
+ * - MNT_OMODE_USER is always used if mount context is in restricted mode
+ * - MNT_OMODE_AUTO is used if nothing other is defined
+ * - the flags are eavaluated in this order: MNT_OMODE_NOTAB, MNT_OMODE_FORCE,
+ *   MNT_OMODE_FSTAB, MNT_OMODE_MTAB and then the mount options from fstab/mtab
+ *   are set according to MNT_OMODE_{IGNORE,APPEND,PREPAND,REPLACE}
  *
  * Returns: 0 on success, negative number in case of error.
  */
 int mnt_context_set_optsmode(struct libmnt_context *cxt, int mode)
 {
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
 	cxt->optsmode = mode;
@@ -216,7 +287,8 @@ int mnt_context_set_optsmode(struct libmnt_context *cxt, int mode)
 
 int mnt_context_get_optsmode(struct libmnt_context *cxt)
 {
-	return cxt ? cxt->optsmode : 0;
+	assert(cxt);
+	return cxt->optsmode;
 }
 
 /**
@@ -228,9 +300,10 @@ int mnt_context_get_optsmode(struct libmnt_context *cxt)
  * canonicalies paths when search in fstab and when prepare source and target paths
  * for mount(2) syscall.
  *
- * This fuction has effect to the private fstab instance only (see
- * mnt_context_set_fstab()). If you want to use an external fstab then you need
- * manage your private struct libmnt_cache (see mnt_table_set_cache(fstab, NULL).
+ * This fuction has effect to the private (within context) fstab instance only
+ * (see mnt_context_set_fstab()). If you want to use an external fstab then you
+ * need manage your private struct libmnt_cache (see mnt_table_set_cache(fstab,
+ * NULL).
  *
  * Returns: 0 on success, negative number in case of error.
  */
@@ -239,6 +312,17 @@ int mnt_context_disable_canonicalize(struct libmnt_context *cxt, int disable)
 	return set_flag(cxt, MNT_FL_NOCANONICALIZE, disable);
 }
 
+/**
+ * mnt_context_is_nocanonicalize:
+ * @cxt: mount context
+ *
+ * Returns: 1 if no-canonicalize mode enabled or 0.
+ */
+int mnt_context_is_nocanonicalize(struct libmnt_context *cxt)
+{
+	return cxt && (cxt->flags & MNT_FL_NOCANONICALIZE) ? 1 : 0;
+}
+
 /**
  * mnt_context_enable_lazy:
  * @cxt: mount context
@@ -261,9 +345,56 @@ int mnt_context_enable_lazy(struct libmnt_context *cxt, int enable)
  */
 int mnt_context_is_lazy(struct libmnt_context *cxt)
 {
-	return cxt && (cxt->flags & MNT_FL_LAZY) ? 1 : 0;
+	return cxt->flags & MNT_FL_LAZY ? 1 : 0;
 }
 
+/**
+ * mnt_context_enable_fork:
+ * @cxt: mount context
+ * @enable: TRUE or FALSE
+ *
+ * Enable/disable fork(2) call in mnt_context_next_mount() (see mount(8) man
+ * page, option -F).
+ *
+ * Returns: 0 on success, negative number in case of error.
+ */
+int mnt_context_enable_fork(struct libmnt_context *cxt, int enable)
+{
+	return set_flag(cxt, MNT_FL_FORK, enable);
+}
+
+/**
+ * mnt_context_is_fork:
+ * @cxt: mount context
+ *
+ * Returns: 1 if fork (mount -F) is enabled or 0
+ */
+int mnt_context_is_fork(struct libmnt_context *cxt)
+{
+	return cxt->flags & MNT_FL_FORK ? 1 : 0;
+}
+
+/**
+ * mnt_context_is_parent:
+ * @cxt: mount context
+ *
+ * Return: 1 if mount -F enabled and the current context is parent, or 0
+ */
+int mnt_context_is_parent(struct libmnt_context *cxt)
+{
+	return mnt_context_is_fork(cxt) && cxt->pid == 0;
+}
+
+/**
+ * mnt_context_is_child:
+ * @cxt: mount context
+ *
+ * Return: 1 if mount -F enabled and the current context is child, or 0
+ */
+int mnt_context_is_child(struct libmnt_context *cxt)
+{
+	return !mnt_context_is_fork(cxt) && cxt->pid;
+}
 
 /**
  * mnt_context_enable_rdonly_umount:
@@ -291,7 +422,7 @@ int mnt_context_enable_rdonly_umount(struct libmnt_context *cxt, int enable)
  */
 int mnt_context_is_rdonly_umount(struct libmnt_context *cxt)
 {
-	return cxt && (cxt->flags & MNT_FL_RDONLY_UMOUNT) ? 1 : 0;
+	return cxt->flags & MNT_FL_RDONLY_UMOUNT ? 1 : 0;
 }
 
 /**
@@ -308,6 +439,18 @@ int mnt_context_disable_helpers(struct libmnt_context *cxt, int disable)
 	return set_flag(cxt, MNT_FL_NOHELPERS, disable);
 }
 
+/**
+ * mnt_context_is_nohelpers
+ * @cxt: mount context
+ *
+ * Returns: 1 if helpers are disabled (mount -i) or 0
+ */
+int mnt_context_is_nohelpers(struct libmnt_context *cxt)
+{
+	return cxt->flags & MNT_FL_NOHELPERS ? 1 : 0;
+}
+
+
 /**
  * mnt_context_enable_sloppy:
  * @cxt: mount context
@@ -330,7 +473,7 @@ int mnt_context_enable_sloppy(struct libmnt_context *cxt, int enable)
  */
 int mnt_context_is_sloppy(struct libmnt_context *cxt)
 {
-	return cxt && (cxt->flags & MNT_FL_SLOPPY) ? 1 : 0;
+	return cxt->flags & MNT_FL_SLOPPY ? 1 : 0;
 }
 
 /**
@@ -355,7 +498,7 @@ int mnt_context_enable_fake(struct libmnt_context *cxt, int enable)
  */
 int mnt_context_is_fake(struct libmnt_context *cxt)
 {
-	return cxt && (cxt->flags & MNT_FL_FAKE) ? 1 : 0;
+	return cxt->flags & MNT_FL_FAKE ? 1 : 0;
 }
 
 /**
@@ -373,14 +516,40 @@ int mnt_context_disable_mtab(struct libmnt_context *cxt, int disable)
 }
 
 /**
- * mnt_context_is_nomtab
+ * mnt_context_is_nomtab:
  * @cxt: mount context
  *
  * Returns: 1 if no-mtab is enabled or 0
  */
 int mnt_context_is_nomtab(struct libmnt_context *cxt)
 {
-	return cxt && (cxt->flags & MNT_FL_NOMTAB) ? 1 : 0;
+	return cxt->flags & MNT_FL_NOMTAB ? 1 : 0;
+}
+
+/**
+ * mnt_context_disable_swapmatch:
+ * @cxt: mount context
+ * @disable: TRUE or FALSE
+ *
+ * Disable/enable swap between source and target for mount(8) if only one path
+ * is specified.
+ *
+ * Returns: 0 on success, negative number in case of error.
+ */
+int mnt_context_disable_swapmatch(struct libmnt_context *cxt, int disable)
+{
+	return set_flag(cxt, MNT_FL_NOSWAPMATCH, disable);
+}
+
+/**
+ * mnt_context_is_swapmatch:
+ * @cxt: mount context
+ *
+ * Returns: 1 if swap between source and target is allowed (default is 1) or 0.
+ */
+int mnt_context_is_swapmatch(struct libmnt_context *cxt)
+{
+	return cxt->flags & MNT_FL_NOSWAPMATCH ? 0 : 1;
 }
 
 /**
@@ -405,7 +574,7 @@ int mnt_context_enable_force(struct libmnt_context *cxt, int enable)
  */
 int mnt_context_is_force(struct libmnt_context *cxt)
 {
-	return cxt && (cxt->flags & MNT_FL_FORCE) ? 1 : 0;
+	return cxt->flags & MNT_FL_FORCE ? 1 : 0;
 }
 
 /**
@@ -430,7 +599,7 @@ int mnt_context_enable_verbose(struct libmnt_context *cxt, int enable)
  */
 int mnt_context_is_verbose(struct libmnt_context *cxt)
 {
-	return cxt && (cxt->flags & MNT_FL_VERBOSE) ? 1 : 0;
+	return cxt->flags & MNT_FL_VERBOSE ? 1 : 0;
 }
 
 /**
@@ -447,6 +616,17 @@ int mnt_context_enable_loopdel(struct libmnt_context *cxt, int enable)
 	return set_flag(cxt, MNT_FL_LOOPDEL, enable);
 }
 
+/**
+ * mnt_context_is_loopdel:
+ * @cxt: mount context
+ *
+ * Returns: 1 if loop device should be deleted after umount (umount -d) or 0.
+ */
+int mnt_context_is_loopdel(struct libmnt_context *cxt)
+{
+	return cxt->flags & MNT_FL_LOOPDEL ? 1 : 0;
+}
+
 /**
  * mnt_context_set_fs:
  * @cxt: mount context
@@ -486,6 +666,7 @@ int mnt_context_set_fs(struct libmnt_context *cxt, struct libmnt_fs *fs)
  */
 struct libmnt_fs *mnt_context_get_fs(struct libmnt_context *cxt)
 {
+	assert(cxt);
 	if (!cxt)
 		return NULL;
 	if (!cxt->fs) {
@@ -504,6 +685,7 @@ struct libmnt_fs *mnt_context_get_fs(struct libmnt_context *cxt)
  */
 int mnt_context_set_source(struct libmnt_context *cxt, const char *source)
 {
+	assert(cxt);
 	return mnt_fs_set_source(mnt_context_get_fs(cxt), source);
 }
 
@@ -515,6 +697,7 @@ int mnt_context_set_source(struct libmnt_context *cxt, const char *source)
  */
 const char *mnt_context_get_source(struct libmnt_context *cxt)
 {
+	assert(cxt);
 	return mnt_fs_get_source(mnt_context_get_fs(cxt));
 }
 
@@ -527,6 +710,7 @@ const char *mnt_context_get_source(struct libmnt_context *cxt)
  */
 int mnt_context_set_target(struct libmnt_context *cxt, const char *target)
 {
+	assert(cxt);
 	return mnt_fs_set_target(mnt_context_get_fs(cxt), target);
 }
 
@@ -538,6 +722,7 @@ int mnt_context_set_target(struct libmnt_context *cxt, const char *target)
  */
 const char *mnt_context_get_target(struct libmnt_context *cxt)
 {
+	assert(cxt);
 	return mnt_fs_get_target(mnt_context_get_fs(cxt));
 }
 
@@ -546,15 +731,15 @@ const char *mnt_context_get_target(struct libmnt_context *cxt)
  * @cxt: mount context
  * @fstype: filesystem type
  *
- * Note that the @fstype has to be the real FS type. For comma-separated list of
- * filesystems or for "nofs" notation use mnt_context_set_fstype_pattern().
+ * Note that the @fstype has to be the real FS type. For patterns with
+ * comma-separated list of filesystems or for "nofs" notation use
+ * mnt_context_set_fstype_pattern().
  *
  * Returns: 0 on success, negative number in case of error.
  */
 int mnt_context_set_fstype(struct libmnt_context *cxt, const char *fstype)
 {
-	if (fstype && strchr(fstype, ','))
-		return -EINVAL;
+	assert(cxt);
 	return mnt_fs_set_fstype(mnt_context_get_fs(cxt), fstype);
 }
 
@@ -562,10 +747,11 @@ int mnt_context_set_fstype(struct libmnt_context *cxt, const char *fstype)
  * mnt_context_get_fstype:
  * @cxt: mount context
  *
- * Returns: returns pointer or NULL in case of error pr if not set.
+ * Returns: pointer or NULL in case of error pr if not set.
  */
 const char *mnt_context_get_fstype(struct libmnt_context *cxt)
 {
+	assert(cxt);
 	return mnt_fs_get_fstype(mnt_context_get_fs(cxt));
 }
 
@@ -578,6 +764,7 @@ const char *mnt_context_get_fstype(struct libmnt_context *cxt)
  */
 int mnt_context_set_options(struct libmnt_context *cxt, const char *optstr)
 {
+	assert(cxt);
 	return mnt_fs_set_options(mnt_context_get_fs(cxt), optstr);
 }
 
@@ -590,9 +777,29 @@ int mnt_context_set_options(struct libmnt_context *cxt, const char *optstr)
  */
 int mnt_context_append_options(struct libmnt_context *cxt, const char *optstr)
 {
+	assert(cxt);
 	return mnt_fs_append_options(mnt_context_get_fs(cxt), optstr);
 }
 
+/**
+ * mnt_context_get_options:
+ * @cxt: mount context
+ *
+ * This function returns mount options set by mnt_context_set_options() or
+ * mnt_context_append_options().
+ *
+ * Note that *after* mnt_context_prepare_mount() may the mount options string
+ * also includes options set by mnt_context_set_mflags() or another options
+ * generated by this library.
+ *
+ * Returns: pointer or NULL
+ */
+const char *mnt_context_get_options(struct libmnt_context *cxt)
+{
+	assert(cxt);
+	return mnt_fs_get_options(mnt_context_get_fs(cxt));
+}
+
 /**
  * mnt_context_set_fstype_pattern:
  * @cxt: mount context
@@ -606,6 +813,7 @@ int mnt_context_set_fstype_pattern(struct libmnt_context *cxt, const char *patte
 {
 	char *p = NULL;
 
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
 	if (pattern) {
@@ -631,6 +839,7 @@ int mnt_context_set_options_pattern(struct libmnt_context *cxt, const char *patt
 {
 	char *p = NULL;
 
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
 	if (pattern) {
@@ -662,6 +871,7 @@ int mnt_context_set_options_pattern(struct libmnt_context *cxt, const char *patt
  */
 int mnt_context_set_fstab(struct libmnt_context *cxt, struct libmnt_table *tb)
 {
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
 	if (!(cxt->flags & MNT_FL_EXTERN_FSTAB))
@@ -685,9 +895,9 @@ int mnt_context_get_fstab(struct libmnt_context *cxt, struct libmnt_table **tb)
 {
 	struct libmnt_cache *cache;
 
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
-
 	if (!cxt->fstab) {
 		int rc;
 
@@ -727,9 +937,9 @@ int mnt_context_get_mtab(struct libmnt_context *cxt, struct libmnt_table **tb)
 {
 	struct libmnt_cache *cache;
 
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
-
 	if (!cxt->mtab) {
 		int rc;
 
@@ -739,6 +949,10 @@ int mnt_context_get_mtab(struct libmnt_context *cxt, struct libmnt_table **tb)
 
 		if (cxt->table_errcb)
 			mnt_table_set_parser_errcb(cxt->mtab, cxt->table_errcb);
+		if (cxt->table_fltrcb)
+			mnt_table_set_parser_fltrcb(cxt->mtab,
+					cxt->table_fltrcb,
+					cxt->table_fltrcb_data);
 
 		rc = mnt_table_parse_mtab(cxt->mtab, cxt->mtab_path);
 		if (rc)
@@ -750,6 +964,33 @@ int mnt_context_get_mtab(struct libmnt_context *cxt, struct libmnt_table **tb)
 
 	if (tb)
 		*tb = cxt->mtab;
+
+	DBG(CXT, mnt_debug_h(cxt, "mtab requested [nents=%d]",
+				mnt_table_get_nents(cxt->mtab)));
+	return 0;
+}
+
+/*
+ * Allows to specify filter for tab file entries. The filter is called by
+ * table parser. Currently used for mtab and utab only.
+ */
+int mnt_context_set_tabfilter(struct libmnt_context *cxt,
+			      int (*fltr)(struct libmnt_fs *, void *),
+			      void *data)
+{
+	assert(cxt);
+	if (!cxt)
+		return -EINVAL;
+
+	cxt->table_fltrcb = fltr;
+	cxt->table_fltrcb_data = data;
+
+	if (cxt->mtab)
+		mnt_table_set_parser_fltrcb(cxt->mtab,
+				cxt->table_fltrcb,
+				cxt->table_fltrcb_data);
+
+	DBG(CXT, mnt_debug_h(cxt, "tabfiler %s", fltr ? "ENABLED!" : "disabled"));
 	return 0;
 }
 
@@ -778,6 +1019,8 @@ int mnt_context_get_table(struct libmnt_context *cxt,
 	struct libmnt_cache *cache;
 	int rc;
 
+	assert(cxt);
+	assert(tb);
 	if (!cxt || !tb)
 		return -EINVAL;
 
@@ -818,6 +1061,7 @@ int mnt_context_get_table(struct libmnt_context *cxt,
 int mnt_context_set_tables_errcb(struct libmnt_context *cxt,
 	int (*cb)(struct libmnt_table *tb, const char *filename, int line))
 {
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
 
@@ -861,7 +1105,8 @@ int mnt_context_set_cache(struct libmnt_context *cxt, struct libmnt_cache *cache
  */
 struct libmnt_cache *mnt_context_get_cache(struct libmnt_context *cxt)
 {
-	if (!cxt || (cxt->flags & MNT_FL_NOCANONICALIZE))
+	assert(cxt);
+	if (!cxt || mnt_context_is_nocanonicalize(cxt))
 		return NULL;
 
 	if (!cxt->cache) {
@@ -873,6 +1118,29 @@ struct libmnt_cache *mnt_context_get_cache(struct libmnt_context *cxt)
 	return cxt->cache;
 }
 
+/**
+ * mnt_context_set_passwd_cb:
+ * @cxt: mount context
+ * @get: callback to get password
+ * @release: callback to release (delallocate) password
+ *
+ * Sets callbacks for encryption password (e.g encrypted loopdev). This
+ * function is deprecated (encrypted loops are no ore supported).
+ *
+ * Returns: 0 on success, negative number in case of error.
+ */
+int mnt_context_set_passwd_cb(struct libmnt_context *cxt,
+			      char *(*get)(struct libmnt_context *),
+			      void (*release)(struct libmnt_context *, char *))
+{
+	assert(cxt);
+	if (!cxt)
+		return -EINVAL;
+	cxt->pwd_get_cb = get;
+	cxt->pwd_release_cb = release;
+	return 0;
+}
+
 /**
  * mnt_context_get_lock:
  * @cxt: mount context
@@ -895,12 +1163,13 @@ struct libmnt_cache *mnt_context_get_cache(struct libmnt_context *cxt)
  */
 struct libmnt_lock *mnt_context_get_lock(struct libmnt_context *cxt)
 {
+	assert(cxt);
 	/*
 	 * DON'T call this function within libmount, it will always allocate
 	 * the lock. The mnt_update_* functions are able to allocate the lock
 	 * only when mtab/utab update is really necessary.
 	 */
-	if (!cxt || (cxt->flags & MNT_FL_NOMTAB))
+	if (!cxt || mnt_context_is_nomtab(cxt))
 		return NULL;
 
 	if (!cxt->lock) {
@@ -934,9 +1203,21 @@ struct libmnt_lock *mnt_context_get_lock(struct libmnt_context *cxt)
  */
 int mnt_context_set_mflags(struct libmnt_context *cxt, unsigned long flags)
 {
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
+
 	cxt->mountflags = flags;
+
+	if ((cxt->flags & MNT_FL_MOUNTOPTS_FIXED) && cxt->fs)
+		/*
+		 * the final mount options are already generated, refresh...
+		 */
+		return mnt_optstr_apply_flags(
+				&cxt->fs->vfs_optstr,
+				cxt->mountflags,
+				mnt_get_builtin_optmap(MNT_LINUX_MAP));
+
 	return 0;
 }
 
@@ -953,6 +1234,10 @@ int mnt_context_set_mflags(struct libmnt_context *cxt, unsigned long flags)
 int mnt_context_get_mflags(struct libmnt_context *cxt, unsigned long *flags)
 {
 	int rc = 0;
+	struct list_head *p;
+
+	assert(cxt);
+	assert(flags);
 	if (!cxt || !flags)
 		return -EINVAL;
 
@@ -963,6 +1248,14 @@ int mnt_context_get_mflags(struct libmnt_context *cxt, unsigned long *flags)
 			rc = mnt_optstr_get_flags(o, flags,
 				    mnt_get_builtin_optmap(MNT_LINUX_MAP));
 	}
+
+	list_for_each(p, &cxt->addmounts) {
+		struct libmnt_addmount *ad =
+				list_entry(p, struct libmnt_addmount, mounts);
+
+		*flags |= ad->mountflags;
+	}
+
 	if (!rc)
 		*flags |= cxt->mountflags;
 	return rc;
@@ -981,6 +1274,7 @@ int mnt_context_get_mflags(struct libmnt_context *cxt, unsigned long *flags)
  */
 int mnt_context_set_user_mflags(struct libmnt_context *cxt, unsigned long flags)
 {
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
 	cxt->user_mountflags = flags;
@@ -1000,6 +1294,9 @@ int mnt_context_set_user_mflags(struct libmnt_context *cxt, unsigned long flags)
 int mnt_context_get_user_mflags(struct libmnt_context *cxt, unsigned long *flags)
 {
 	int rc = 0;
+
+	assert(cxt);
+	assert(flags);
 	if (!cxt || !flags)
 		return -EINVAL;
 
@@ -1031,6 +1328,7 @@ int mnt_context_get_user_mflags(struct libmnt_context *cxt, unsigned long *flags
  */
 int mnt_context_set_mountdata(struct libmnt_context *cxt, void *data)
 {
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
 	cxt->mountdata = data;
@@ -1059,10 +1357,14 @@ int mnt_context_prepare_srcpath(struct libmnt_context *cxt)
 
 	src = mnt_fs_get_source(cxt->fs);
 
+	if (!src && mnt_context_propagation_only(cxt))
+		/* mount --make-{shared,private,...} */
+		return mnt_fs_set_source(cxt->fs, "none");
+
 	/* ignore filesystems without source or filesystems
 	 * where the source is quasi-path (//foo/bar)
 	 */
-	if (!src || (cxt->fs->flags & MNT_FS_NET))
+	if (!src || mnt_fs_is_netfs(cxt->fs))
 		return 0;
 
 	DBG(CXT, mnt_debug_h(cxt, "srcpath '%s'", src));
@@ -1076,9 +1378,9 @@ int mnt_context_prepare_srcpath(struct libmnt_context *cxt)
 		if (cache)
 			path = mnt_resolve_tag(t, v, cache);
 
-		rc = path ? mnt_fs_set_source(cxt->fs, path) : -EINVAL;
+		rc = path ? mnt_fs_set_source(cxt->fs, path) : -MNT_ERR_NOSOURCE;
 
-	} else if (cache) {
+	} else if (cache && !mnt_fs_is_pseudofs(cxt->fs)) {
 		/*
 		 * Source is PATH (canonicalize)
 		 */
@@ -1095,9 +1397,9 @@ int mnt_context_prepare_srcpath(struct libmnt_context *cxt)
 	if (!path)
 		path = src;
 
-	if ((cxt->mountflags & (MS_BIND | MS_MOVE | MS_PROPAGATION)) ||
-	    (cxt->fs->flags & MNT_FS_PSEUDO)) {
-		DBG(CXT, mnt_debug_h(cxt, "PROPAGATION/pseudo FS source: %s", path));
+	if ((cxt->mountflags & (MS_BIND | MS_MOVE | MS_REMOUNT))
+	    || mnt_fs_is_pseudofs(cxt->fs)) {
+		DBG(CXT, mnt_debug_h(cxt, "REMOUNT/BIND/MOVE/pseudo FS source: %s", path));
 		return rc;
 	}
 
@@ -1115,6 +1417,47 @@ int mnt_context_prepare_srcpath(struct libmnt_context *cxt)
 	return 0;
 }
 
+/* create a mountpoint if x-mount.mkdir[=<mode>] specified */
+static int mkdir_target(const char *tgt, struct libmnt_fs *fs)
+{
+	char *mstr = NULL;
+	size_t mstr_sz = 0;
+	mode_t mode = 0;
+	struct stat st;
+	int rc;
+
+	assert(tgt);
+	assert(fs);
+
+	if (mnt_optstr_get_option(fs->user_optstr, "x-mount.mkdir", &mstr, &mstr_sz) != 0)
+		return 0;
+	if (stat(tgt, &st) == 0)
+		return 0;
+
+	if (mstr && mstr_sz) {
+		char *end = NULL;
+
+		errno = 0;
+		mode = strtol(mstr, &end, 8);
+
+		if (errno || !end || mstr + mstr_sz != end) {
+			DBG(CXT, mnt_debug("failed to parse mkdir mode '%s'", mstr));
+			return -MNT_ERR_MOUNTOPT;
+		}
+	}
+
+	if (!mode)
+		mode = S_IRWXU |			/* 0755 */
+		       S_IRGRP | S_IXGRP |
+		       S_IROTH | S_IXOTH;
+
+	rc = mkdir_p(tgt, mode);
+	if (rc)
+		DBG(CXT, mnt_debug("mkdir %s failed: %m", tgt));
+
+	return rc;
+}
+
 int mnt_context_prepare_target(struct libmnt_context *cxt)
 {
 	const char *tgt;
@@ -1134,26 +1477,41 @@ int mnt_context_prepare_target(struct libmnt_context *cxt)
 	if (!tgt)
 		return 0;
 
+	/* mkdir target */
+	if (cxt->action == MNT_ACT_MOUNT
+	    && !mnt_context_is_restricted(cxt)
+	    && cxt->user_mountflags & MNT_MS_XCOMMENT) {
+
+		rc = mkdir_target(tgt, cxt->fs);
+		if (rc)
+			return rc;	/* mkdir or parse error */
+	}
+
+	/* canonicalize the path */
 	cache = mnt_context_get_cache(cxt);
 	if (cache) {
 		char *path = mnt_resolve_path(tgt, cache);
-		if (strcmp(path, tgt))
+		if (path && strcmp(path, tgt) != 0)
 			rc = mnt_fs_set_target(cxt->fs, path);
 	}
 
 	if (rc)
-		DBG(CXT, mnt_debug_h(cxt, "failed to prepare target"));
+		DBG(CXT, mnt_debug_h(cxt, "failed to prepare target '%s'", tgt));
 	else
 		DBG(CXT, mnt_debug_h(cxt, "final target '%s'",
 					mnt_fs_get_target(cxt->fs)));
 	return 0;
 }
 
+/*
+ * It's usully no error when we're not able to detect filesystem type -- we
+ * will try to use types from /{etc,proc}/filesystems.
+ */
 int mnt_context_guess_fstype(struct libmnt_context *cxt)
 {
 	char *type;
 	const char *dev;
-	int rc = -EINVAL;
+	int rc = 0;
 
 	assert(cxt);
 	assert(cxt->fs);
@@ -1162,7 +1520,8 @@ int mnt_context_guess_fstype(struct libmnt_context *cxt)
 	if (!cxt || !cxt->fs)
 		return -EINVAL;
 
-	if (cxt->mountflags & (MS_BIND | MS_MOVE | MS_PROPAGATION))
+	if ((cxt->mountflags & (MS_BIND | MS_MOVE))
+	    || mnt_context_propagation_only(cxt))
 		goto none;
 
 	type = (char *) mnt_fs_get_fstype(cxt->fs);
@@ -1180,34 +1539,34 @@ int mnt_context_guess_fstype(struct libmnt_context *cxt)
 
 	dev = mnt_fs_get_srcpath(cxt->fs);
 	if (!dev)
-		goto err;
+		goto done;
 
 	if (access(dev, F_OK) == 0) {
 		struct libmnt_cache *cache = mnt_context_get_cache(cxt);
+		int ambi = 0;
 
-		type = mnt_get_fstype(dev, &cxt->ambi, cache);
+		type = mnt_get_fstype(dev, &ambi, cache);
 		if (type) {
 			rc = mnt_fs_set_fstype(cxt->fs, type);
 			if (!cache)
 				free(type);	/* type is not cached */
 		}
+		if (ambi)
+			rc = -MNT_ERR_AMBIFS;
 	} else {
+		DBG(CXT, mnt_debug_h(cxt, "access(%s) failed [%m]", dev));
 		if (strchr(dev, ':') != NULL)
 			rc = mnt_fs_set_fstype(cxt->fs, "nfs");
 		else if (!strncmp(dev, "//", 2))
 			rc = mnt_fs_set_fstype(cxt->fs, "cifs");
 	}
-	if (rc)
-		goto err;
+
 done:
-	DBG(CXT, mnt_debug_h(cxt, "FS type: %s",
-				mnt_fs_get_fstype(cxt->fs)));
-	return 0;
+	DBG(CXT, mnt_debug_h(cxt, "FS type: %s [rc=%d]",
+				mnt_fs_get_fstype(cxt->fs), rc));
+	return rc;
 none:
 	return mnt_fs_set_fstype(cxt->fs, "none");
-err:
-	DBG(CXT, mnt_debug_h(cxt, "failed to detect FS type"));
-	return rc;
 }
 
 /*
@@ -1230,8 +1589,13 @@ int mnt_context_prepare_helper(struct libmnt_context *cxt, const char *name,
 	if (!type)
 		type = mnt_fs_get_fstype(cxt->fs);
 
-	if ((cxt->flags & MNT_FL_NOHELPERS) || !type ||
-	    !strcmp(type, "none") || (cxt->fs->flags & MNT_FS_SWAP))
+	if (type && strchr(type, ','))
+		return 0;			/* type is fstype pattern */
+
+	if (mnt_context_is_nohelpers(cxt)
+	    || !type
+	    || !strcmp(type, "none")
+	    || mnt_fs_is_swaparea(cxt->fs))
 		return 0;
 
 	path = strtok_r(search_path, ":", &p);
@@ -1250,7 +1614,9 @@ int mnt_context_prepare_helper(struct libmnt_context *cxt, const char *name,
 		rc = stat(helper, &st);
 		if (rc == -1 && errno == ENOENT && strchr(type, '.')) {
 			/* If type ends with ".subtype" try without it */
-			*strrchr(helper, '.') = '\0';
+			char *hs = strrchr(helper, '.');
+			if (hs)
+				*hs = '\0';
 			rc = stat(helper, &st);
 		}
 
@@ -1283,11 +1649,6 @@ int mnt_context_merge_mflags(struct libmnt_context *cxt)
 		return rc;
 	cxt->mountflags = fl;
 
-	/* TODO: if cxt->fs->fs_optstr contains 'ro' then set the MS_RDONLY to
-	 * mount flags, it's possible that superblock is read-only, but VFS is
-	 * read-write.
-	 */
-
 	fl = 0;
 	rc = mnt_context_get_user_mflags(cxt, &fl);
 	if (rc)
@@ -1316,8 +1677,8 @@ int mnt_context_prepare_update(struct libmnt_context *cxt)
 
 	DBG(CXT, mnt_debug_h(cxt, "prepare update"));
 
-	if (cxt->mountflags & MS_PROPAGATION) {
-		DBG(CXT, mnt_debug_h(cxt, "skip update: MS_PROPAGATION"));
+	if (mnt_context_propagation_only(cxt)) {
+		DBG(CXT, mnt_debug_h(cxt, "skip update: only MS_PROPAGATION"));
 		return 0;
 	}
 
@@ -1325,9 +1686,9 @@ int mnt_context_prepare_update(struct libmnt_context *cxt)
 
 	if (cxt->action == MNT_ACT_UMOUNT && target && !strcmp(target, "/"))
 		/* Don't try to touch mtab if umounting root FS */
-		cxt->flags |= MNT_FL_NOMTAB;
+		mnt_context_disable_mtab(cxt, TRUE);
 
-	if (cxt->flags & MNT_FL_NOMTAB) {
+	if (mnt_context_is_nomtab(cxt)) {
 		DBG(CXT, mnt_debug_h(cxt, "skip update: NOMTAB flag"));
 		return 0;
 	}
@@ -1346,19 +1707,26 @@ int mnt_context_prepare_update(struct libmnt_context *cxt)
 				cxt->syscall_status));
 		return 0;
 	}
+
 	if (!cxt->update) {
+		const char *name = cxt->mtab_writable ? cxt->mtab_path : cxt->utab_path;
+
+		if (cxt->action == MNT_ACT_UMOUNT && is_file_empty(name)) {
+			DBG(CXT, mnt_debug_h(cxt,
+				"skip update: umount, no table"));
+			return 0;
+		}
+
 		cxt->update = mnt_new_update();
 		if (!cxt->update)
 			return -ENOMEM;
 
-		mnt_update_set_filename(cxt->update,
-				cxt->mtab_writable ? cxt->mtab_path : cxt->utab_path,
-				!cxt->mtab_writable);
+		mnt_update_set_filename(cxt->update, name, !cxt->mtab_writable);
 	}
 
 	if (cxt->action == MNT_ACT_UMOUNT)
 		rc = mnt_update_set_fs(cxt->update, cxt->mountflags,
-					mnt_fs_get_target(cxt->fs), NULL);
+					mnt_context_get_target(cxt), NULL);
 	else
 		rc = mnt_update_set_fs(cxt->update, cxt->mountflags,
 					NULL, cxt->fs);
@@ -1372,7 +1740,7 @@ int mnt_context_update_tabs(struct libmnt_context *cxt)
 
 	assert(cxt);
 
-	if (cxt->flags & MNT_FL_NOMTAB) {
+	if (mnt_context_is_nomtab(cxt)) {
 		DBG(CXT, mnt_debug_h(cxt, "don't update: NOMTAB flag"));
 		return 0;
 	}
@@ -1424,7 +1792,7 @@ static int apply_table(struct libmnt_context *cxt, struct libmnt_table *tb,
 		else if (tgt)
 			fs = mnt_table_find_target(tb, tgt, direction);
 
-		if (!fs) {
+		if (!fs && mnt_context_is_swapmatch(cxt)) {
 			/* swap source and target (if @src is not LABEL/UUID),
 			 * for example in
 			 *
@@ -1441,7 +1809,7 @@ static int apply_table(struct libmnt_context *cxt, struct libmnt_table *tb,
 	}
 
 	if (!fs)
-		return -EINVAL;
+		return -MNT_ERR_NOFSTAB;	/* not found */
 
 	DBG(CXT, mnt_debug_h(cxt, "apply entry:"));
 	DBG(CXT, mnt_fs_print_debug(fs, stderr));
@@ -1494,7 +1862,7 @@ int mnt_context_apply_fstab(struct libmnt_context *cxt)
 	if (!cxt)
 		return -EINVAL;
 
-	if (cxt->flags & MNT_FL_TAB_APPLIED)
+	if (mnt_context_tab_applied(cxt))	/* already applied */
 		return 0;
 
 	if (mnt_context_is_restricted(cxt)) {
@@ -1503,7 +1871,10 @@ int mnt_context_apply_fstab(struct libmnt_context *cxt)
 	} else if (cxt->optsmode == 0) {
 		DBG(CXT, mnt_debug_h(cxt, "use default optmode"));
 		cxt->optsmode = MNT_OMODE_AUTO;
-
+	} else if (cxt->optsmode & MNT_OMODE_NOTAB) {
+		cxt->optsmode &= ~MNT_OMODE_FSTAB;
+		cxt->optsmode &= ~MNT_OMODE_MTAB;
+		cxt->optsmode &= ~MNT_OMODE_FORCE;
 	}
 
 	if (cxt->fs) {
@@ -1527,11 +1898,19 @@ int mnt_context_apply_fstab(struct libmnt_context *cxt)
 		return 0;
 	}
 
+	if (!src && tgt
+	    && !(cxt->optsmode & MNT_OMODE_FSTAB)
+	    && !(cxt->optsmode & MNT_OMODE_MTAB)) {
+		DBG(CXT, mnt_debug_h(cxt, "only target; fstab/mtab not required "
+					  "-- skip, probably MS_PROPAGATION"));
+		return 0;
+	}
+
 	DBG(CXT, mnt_debug_h(cxt,
 		"trying to apply fstab (src=%s, target=%s)", src, tgt));
 
 	/* let's initialize cxt->fs */
-	mnt_context_get_fs(cxt);
+	ignore_result( mnt_context_get_fs(cxt) );
 
 	/* try fstab */
 	if (cxt->optsmode & MNT_OMODE_FSTAB) {
@@ -1542,7 +1921,7 @@ int mnt_context_apply_fstab(struct libmnt_context *cxt)
 
 	/* try mtab */
 	if (rc < 0 && (cxt->optsmode & MNT_OMODE_MTAB)) {
-		DBG(CXT, mnt_debug_h(cxt, "tring to apply from mtab"));
+		DBG(CXT, mnt_debug_h(cxt, "trying to apply from mtab"));
 		rc = mnt_context_get_mtab(cxt, &tab);
 		if (!rc)
 			rc = apply_table(cxt, tab, MNT_ITER_BACKWARD);
@@ -1552,15 +1931,112 @@ int mnt_context_apply_fstab(struct libmnt_context *cxt)
 	return rc;
 }
 
+/**
+ * mnt_context_tab_applied:
+ * @cxt: mount context
+ *
+ * Returns: 1 if fstab (or mtab) has been applied to the context, or 0.
+ */
+int mnt_context_tab_applied(struct libmnt_context *cxt)
+{
+	assert(cxt);
+	return cxt->flags & MNT_FL_TAB_APPLIED;
+}
+
+/*
+ * This is not public function!
+ *
+ * Returns 1 if *only propagation flags* change is requested.
+ */
+int mnt_context_propagation_only(struct libmnt_context *cxt)
+{
+	assert(cxt);
+	assert(cxt->fs);
+
+	if (cxt->action != MNT_ACT_MOUNT)
+		return 0;
+
+	/* has to be called after context_mount.c: fix_opts() */
+	assert((cxt->flags & MNT_FL_MOUNTOPTS_FIXED));
+
+	/* all propagation mounts are in cxt->addmount */
+	return !list_empty(&cxt->addmounts)
+	       && (cxt->mountflags == 0 || cxt->mountflags == MS_SILENT)
+	       && cxt->fs
+	       && (!cxt->fs->fstype || strcmp(cxt->fs->fstype, "none") == 0)
+	       && (!cxt->fs->source || strcmp(cxt->fs->source, "none") == 0);
+}
+
 /**
  * mnt_context_get_status:
  * @cxt: mount context
  *
- * Returns: 1 if /sbin/mount.type or mount(2) syscall was successfull.
+ * Global libmount status.
+ *
+ * The real exit code of the mount.type helper has to be tested by
+ * mnt_context_get_helper_status(). The mnt_context_get_status() only inform
+ * that exec() has been successful.
+ *
+ * Returns: 1 if mount.type or mount(2) syscall has been successfully called.
  */
 int mnt_context_get_status(struct libmnt_context *cxt)
 {
-	return cxt && (!cxt->syscall_status || !cxt->helper_exec_status);
+	assert(cxt);
+	return !cxt->syscall_status || !cxt->helper_exec_status;
+}
+
+/**
+ * mnt_context_helper_executed:
+ * @cxt: mount context
+ *
+ * Returns: 1 if mount.type helper has been executed, or 0.
+ */
+int mnt_context_helper_executed(struct libmnt_context *cxt)
+{
+	assert(cxt);
+	return cxt->helper_exec_status != 1;
+}
+
+/**
+ * mnt_context_get_helper_status:
+ * @cxt: mount context
+ *
+ * Return: mount.type helper exit status, result is reliable only if
+ *         mnt_context_helper_executed() returns 1.
+ */
+int mnt_context_get_helper_status(struct libmnt_context *cxt)
+{
+	assert(cxt);
+	return cxt->helper_status;
+}
+
+/**
+ * mnt_context_syscall_called:
+ * @cxt: mount context
+ *
+ * Returns: 1 if mount(2) syscall has been called, or 0.
+ */
+int mnt_context_syscall_called(struct libmnt_context *cxt)
+{
+	assert(cxt);
+	return cxt->syscall_status != 1;
+}
+
+/**
+ * mnt_context_get_syscall_errno:
+ * @cxt: mount context
+ *
+ * The result from this function is reliable only if
+ * mnt_context_syscall_called() returns 1.
+ *
+ * Returns: mount(2) errno if the syscall failed or 0.
+ */
+int mnt_context_get_syscall_errno(struct libmnt_context *cxt)
+{
+	assert(cxt);
+	if (cxt->syscall_status < 0)
+		return -cxt->syscall_status;
+	return 0;
 }
 
 /**
@@ -1568,16 +2044,16 @@ int mnt_context_get_status(struct libmnt_context *cxt)
  * @cxt: mount context
  * @status: mount(2) status
  *
- * The @status should be 0 on succcess, or negative number on error (-1 or
- * -errno).
+ * The @status should be 0 on success, or negative number on error (-errno).
  *
- * This function should be used if [u]mount(2) syscall was NOT called by
- * libmount (by mnt_context_mount() or mnt_context_do_mount()) only.
+ * This function should be used only if [u]mount(2) syscall is NOT called by
+ * libmount code.
  *
  * Returns: 0 or negative number in case of error.
  */
 int mnt_context_set_syscall_status(struct libmnt_context *cxt, int status)
 {
+	assert(cxt);
 	if (!cxt)
 		return -EINVAL;
 
@@ -1592,6 +2068,8 @@ int mnt_context_set_syscall_status(struct libmnt_context *cxt, int status)
  * @buf: buffer
  * @bufsiz: size of the buffer
  *
+ * Not implemented yet.
+ *
  * Returns: 0 or negative number in case of error.
  */
 int mnt_context_strerror(struct libmnt_context *cxt __attribute__((__unused__)),
@@ -1622,8 +2100,11 @@ int mnt_context_strerror(struct libmnt_context *cxt __attribute__((__unused__)),
 int mnt_context_init_helper(struct libmnt_context *cxt, int action,
 			    int flags __attribute__((__unused__)))
 {
-	int rc = mnt_context_disable_helpers(cxt, TRUE);
+	int rc;
 
+	assert(cxt);
+
+	rc = mnt_context_disable_helpers(cxt, TRUE);
 	if (!rc)
 		rc = set_flag(cxt, MNT_FL_HELPER, 1);
 	if (!rc)
@@ -1664,6 +2145,8 @@ int mnt_context_helper_setopt(struct libmnt_context *cxt, int c, char *arg)
  * @fs: filesystem
  * @mounted: returns 1 for mounted and 0 for non-mounted filesystems
  *
+ * Please, read mnt_table_is_fs_mounted() description!
+ *
  * Returns: 0 on success and negative number in case of error.
  */
 int mnt_context_is_fs_mounted(struct libmnt_context *cxt,
@@ -1672,6 +2155,7 @@ int mnt_context_is_fs_mounted(struct libmnt_context *cxt,
 	struct libmnt_table *mtab;
 	int rc;
 
+	assert(cxt);
 	if (!cxt || !fs || !mounted)
 		return -EINVAL;
 
@@ -1683,6 +2167,105 @@ int mnt_context_is_fs_mounted(struct libmnt_context *cxt,
 	return 0;
 }
 
+static int mnt_context_add_child(struct libmnt_context *cxt, pid_t pid)
+{
+	pid_t *pids;
+
+	assert(cxt);
+	if (!cxt)
+		return -EINVAL;
+
+	pids = realloc(cxt->children, sizeof(pid_t) * cxt->nchildren + 1);
+	if (!pids)
+		return -ENOMEM;
+
+	DBG(CXT, mnt_debug_h(cxt, "add new child %d", pid));
+	cxt->children = pids;
+	cxt->children[cxt->nchildren++] = pid;
+
+	return 0;
+}
+
+int mnt_fork_context(struct libmnt_context *cxt)
+{
+	int rc = 0;
+	pid_t pid;
+
+	assert(cxt);
+	if (!mnt_context_is_parent(cxt))
+		return -EINVAL;
+
+	DBG(CXT, mnt_debug_h(cxt, "forking context"));
+
+	DBG_FLUSH;
+
+	pid = fork();
+
+	switch (pid) {
+	case -1: /* error */
+		DBG(CXT, mnt_debug_h(cxt, "fork failed %m"));
+		return -errno;
+
+	case 0: /* child */
+		cxt->pid = getpid();
+		mnt_context_enable_fork(cxt, FALSE);
+		DBG(CXT, mnt_debug_h(cxt, "child created"));
+		break;
+
+	default:
+		rc = mnt_context_add_child(cxt, pid);
+		break;
+	}
+
+	return rc;
+}
+
+int mnt_context_wait_for_children(struct libmnt_context *cxt,
+				  int *nchildren, int *nerrs)
+{
+	int i;
+
+	assert(cxt);
+	if (!cxt)
+		return -EINVAL;
+
+	assert(mnt_context_is_parent(cxt));
+
+	for (i = 0; i < cxt->nchildren; i++) {
+		pid_t pid = cxt->children[i];
+		int rc = 0, ret = 0;
+
+		if (!pid)
+			continue;
+		do {
+			DBG(CXT, mnt_debug_h(cxt,
+					"waiting for child (%d/%d): %d",
+					i + 1, cxt->nchildren, pid));
+			errno = 0;
+			rc = waitpid(pid, &ret, 0);
+
+		} while (rc == -1 && errno == EINTR);
+
+		if (nchildren)
+			(*nchildren)++;
+
+		if (rc != -1 && nerrs) {
+			if (WIFEXITED(ret))
+				(*nerrs) += WEXITSTATUS(ret) == 0 ? 0 : 1;
+			else
+				(*nerrs)++;
+		}
+		cxt->children[i] = 0;
+	}
+
+	cxt->nchildren = 0;
+	free(cxt->children);
+	cxt->children = NULL;
+	return 0;
+}
+
+
+
 #ifdef TEST_PROGRAM
 
 struct libmnt_lock *lock;
@@ -1725,7 +2308,7 @@ int test_mount(struct libmnt_test *ts, int argc, char *argv[])
 		mnt_context_set_target(cxt, argv[idx++]);
 	}
 
-	/* this is unnecessary -- libmount is able to internaly
+	/* this is unnecessary! -- libmount is able to internaly
 	 * create and manage the lock
 	 */
 	lock = mnt_context_get_lock(cxt);
@@ -1734,10 +2317,11 @@ int test_mount(struct libmnt_test *ts, int argc, char *argv[])
 
 	rc = mnt_context_mount(cxt);
 	if (rc)
-		printf("failed to mount %s\n", strerror(errno));
+		printf("failed to mount: %m\n");
 	else
 		printf("successfully mounted\n");
 
+	lock = NULL;	/* because we use atexit lock_fallback */
 	mnt_free_context(cxt);
 	return rc;
 }
@@ -1792,6 +2376,7 @@ int test_umount(struct libmnt_test *ts, int argc, char *argv[])
 	else
 		printf("successfully umounted\n");
 err:
+	lock = NULL;	/* because we use atexit lock_fallback */
 	mnt_free_context(cxt);
 	return rc;
 }
@@ -1848,11 +2433,11 @@ int test_mountall(struct libmnt_test *ts, int argc, char *argv[])
 		return -ENOMEM;
 
 	if (argc > 2) {
-		if (!strcmp(argv[idx], "-O")) {
+		if (argv[idx] && !strcmp(argv[idx], "-O")) {
 			mnt_context_set_options_pattern(cxt, argv[idx + 1]);
 			idx += 2;
 		}
-		if (!strcmp(argv[idx], "-t")) {
+		if (argv[idx] && !strcmp(argv[idx], "-t")) {
 			mnt_context_set_fstype_pattern(cxt, argv[idx + 1]);
 			idx += 2;
 		}
