@@ -14,9 +14,9 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
- * You should have received a copy of the GNU General Public License
- * along with this program; if not, write to the Free Software Foundation,
- * Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+ * You should have received a copy of the GNU General Public License along
+ * with this program; if not, write to the Free Software Foundation, Inc.,
+ * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 #include <stdio.h>
 #include <stdlib.h>
@@ -26,18 +26,26 @@
 #include <string.h>
 #include <termios.h>
 #ifdef HAVE_SYS_IOCTL_H
-#include <sys/ioctl.h>
+# include <sys/ioctl.h>
 #endif
 #include <assert.h>
 #include <poll.h>
-
+#include <sys/statvfs.h>
+#include <sys/types.h>
+#ifdef HAVE_LIBUDEV
+# include <libudev.h>
+#endif
 #include <libmount.h>
 
 #include "pathnames.h"
 #include "nls.h"
+#include "closestream.h"
 #include "c.h"
 #include "tt.h"
 #include "strutils.h"
+#include "xalloc.h"
+#include "optutils.h"
+#include "mangle.h"
 
 /* flags */
 enum {
@@ -48,7 +56,9 @@ enum {
 	FL_NOSWAPMATCH	= (1 << 6),
 	FL_NOFSROOT	= (1 << 7),
 	FL_SUBMOUNTS	= (1 << 8),
-	FL_POLL		= (1 << 9)
+	FL_POLL		= (1 << 9),
+	FL_DF		= (1 << 10),
+	FL_ALL		= (1 << 11)
 };
 
 /* column IDs */
@@ -61,14 +71,33 @@ enum {
 	COL_FS_OPTIONS,
 	COL_LABEL,
 	COL_UUID,
+	COL_PARTLABEL,
+	COL_PARTUUID,
 	COL_MAJMIN,
 	COL_ACTION,
 	COL_OLD_TARGET,
 	COL_OLD_OPTIONS,
+	COL_SIZE,
+	COL_AVAIL,
+	COL_USED,
+	COL_USEPERC,
+	COL_FSROOT,
+	COL_TID,
+	COL_ID,
+	COL_OPT_FIELDS,
+	COL_PROPAGATION,
+	COL_FREQ,
+	COL_PASSNO,
 
 	FINDMNT_NCOLUMNS
 };
 
+enum {
+	TABTYPE_FSTAB = 1,
+	TABTYPE_MTAB,
+	TABTYPE_KERNEL
+};
+
 /* column names */
 struct colinfo {
 	const char	*name;		/* header */
@@ -76,22 +105,36 @@ struct colinfo {
 	int		flags;		/* tt flags */
 	const char      *help;		/* column description */
 	const char	*match;		/* pattern for match_func() */
+	void		*match_data;	/* match specific data */
 };
 
 /* columns descriptions (don't use const, this is writable) */
 static struct colinfo infos[FINDMNT_NCOLUMNS] = {
-	[COL_SOURCE]       = { "SOURCE",       0.25, 0, N_("source device") },
-	[COL_TARGET]       = { "TARGET",       0.30, TT_FL_TREE, N_("mountpoint") },
+	[COL_SOURCE]       = { "SOURCE",       0.25, TT_FL_NOEXTREMES, N_("source device") },
+	[COL_TARGET]       = { "TARGET",       0.30, TT_FL_TREE | TT_FL_NOEXTREMES, N_("mountpoint") },
 	[COL_FSTYPE]       = { "FSTYPE",       0.10, TT_FL_TRUNC, N_("filesystem type") },
 	[COL_OPTIONS]      = { "OPTIONS",      0.10, TT_FL_TRUNC, N_("all mount options") },
 	[COL_VFS_OPTIONS]  = { "VFS-OPTIONS",  0.20, TT_FL_TRUNC, N_("VFS specific mount options") },
 	[COL_FS_OPTIONS]   = { "FS-OPTIONS",   0.10, TT_FL_TRUNC, N_("FS specific mount options") },
 	[COL_LABEL]        = { "LABEL",        0.10, 0, N_("filesystem label") },
 	[COL_UUID]         = { "UUID",           36, 0, N_("filesystem UUID") },
+	[COL_PARTLABEL]    = { "PARTLABEL",    0.10, 0, N_("partition label") },
+	[COL_PARTUUID]     = { "PARTUUID",       36, 0, N_("partition UUID") },
 	[COL_MAJMIN]       = { "MAJ:MIN",         6, 0, N_("major:minor device number") },
 	[COL_ACTION]       = { "ACTION",         10, TT_FL_STRICTWIDTH, N_("action detected by --poll") },
 	[COL_OLD_OPTIONS]  = { "OLD-OPTIONS",  0.10, TT_FL_TRUNC, N_("old mount options saved by --poll") },
 	[COL_OLD_TARGET]   = { "OLD-TARGET",   0.30, 0, N_("old mountpoint saved by --poll") },
+	[COL_SIZE]         = { "SIZE",            5, TT_FL_RIGHT, N_("filesystem size") },
+	[COL_AVAIL]        = { "AVAIL",           5, TT_FL_RIGHT, N_("filesystem size available") },
+	[COL_USED]         = { "USED",            5, TT_FL_RIGHT, N_("filesystem size used") },
+	[COL_USEPERC]      = { "USE%",            3, TT_FL_RIGHT, N_("filesystem use percentage") },
+	[COL_FSROOT]       = { "FSROOT",       0.25, TT_FL_NOEXTREMES, N_("filesystem root") },
+	[COL_TID]          = { "TID",             4, TT_FL_RIGHT, N_("task ID") },
+	[COL_ID]           = { "ID",              2, TT_FL_RIGHT, N_("mount ID") },
+	[COL_OPT_FIELDS]   = { "OPT-FIELDS",   0.10, TT_FL_TRUNC, N_("optional mount fields") },
+	[COL_PROPAGATION]  = { "PROPAGATION",  0.10, 0, N_("VFS propagation flags") },
+	[COL_FREQ]         = { "FREQ",            1, TT_FL_RIGHT, N_("dump(8) frequency in days [fstab only]") },
+	[COL_PASSNO]       = { "PASSNO",          1, TT_FL_RIGHT, N_("pass number on parallel fsck(8) [fstab only]") }
 };
 
 /* global flags */
@@ -110,6 +153,13 @@ static int nactions;
 /* libmount cache */
 static struct libmnt_cache *cache;
 
+#ifdef HAVE_LIBUDEV
+struct udev *udev;
+#endif
+
+static int match_func(struct libmnt_fs *fs, void *data __attribute__ ((__unused__)));
+
+
 static int get_column_id(int num)
 {
 	assert(num < ncolumns);
@@ -149,12 +199,64 @@ static const char *get_match(int id)
 	return infos[id].match;
 }
 
+static void *get_match_data(int id)
+{
+	assert(id < FINDMNT_NCOLUMNS);
+	return infos[id].match_data;
+}
+
 static void set_match(int id, const char *match)
 {
 	assert(id < FINDMNT_NCOLUMNS);
 	infos[id].match = match;
 }
 
+static void set_match_data(int id, void *data)
+{
+	assert(id < FINDMNT_NCOLUMNS);
+	infos[id].match_data = data;
+}
+
+/*
+ * source match means COL_SOURCE *or* COL_MAJMIN, depends on
+ * data format.
+ */
+static void set_source_match(const char *data)
+{
+	int maj, min;
+
+	if (sscanf(data, "%d:%d", &maj, &min) == 2) {
+		dev_t *devno = xmalloc(sizeof(dev_t));
+
+		*devno = makedev(maj, min);
+		set_match(COL_MAJMIN, data);
+		set_match_data(COL_MAJMIN, (void *) devno);
+		flags |= FL_NOSWAPMATCH;
+	} else
+		set_match(COL_SOURCE, data);
+}
+
+static void enable_extra_target_match(void)
+{
+	char *cn = NULL, *mnt = NULL;
+
+	/*
+	 * Check if match pattern is mountpoint, if not use the
+	 * real mountpoint.
+	 */
+	cn = mnt_resolve_path(get_match(COL_TARGET), cache);
+	if (!cn)
+		return;
+
+	mnt = mnt_get_mountpoint(cn);
+	if (!mnt || strcmp(mnt, cn) == 0)
+		return;
+
+	/* replace the current setting with the real mountpoint */
+	set_match(COL_TARGET, mnt);
+}
+
+
 static int is_tabdiff_column(int id)
 {
 	assert(id < FINDMNT_NCOLUMNS);
@@ -175,10 +277,14 @@ static int is_tabdiff_column(int id)
  */
 static int is_listall_mode(void)
 {
+	if ((flags & FL_DF) && !(flags & FL_ALL))
+		return 0;
+
 	return (!get_match(COL_SOURCE) &&
 		!get_match(COL_TARGET) &&
 		!get_match(COL_FSTYPE) &&
-		!get_match(COL_OPTIONS));
+		!get_match(COL_OPTIONS) &&
+		!get_match(COL_MAJMIN));
 }
 
 /*
@@ -256,32 +362,129 @@ static int column_name_to_id(const char *name, size_t namesz)
 	return -1;
 }
 
-/* Returns LABEL or UUID */
-static const char *get_tag(struct libmnt_fs *fs, const char *tagname)
+
+#ifdef HAVE_LIBUDEV
+static char *get_tag_from_udev(const char *devname, int col)
 {
-	const char *t, *v, *res;
+	struct udev_device *dev;
+	const char *data = NULL;
+	char *res = NULL, *path;
+
+	if (!udev)
+		udev = udev_new();
+	if (!udev)
+		return NULL;
+
+	/* libudev don't like /dev/mapper/ symlinks */
+	path = realpath(devname, NULL);
+	if (path)
+		devname = path;
+
+	if (strncmp(devname, "/dev/", 5) == 0)
+		devname += 5;
+
+	dev = udev_device_new_from_subsystem_sysname(udev, "block", devname);
+	free(path);
+
+	if (!dev)
+		return NULL;
+
+	switch(col) {
+	case COL_LABEL:
+		data = udev_device_get_property_value(dev, "ID_FS_LABEL_ENC");
+		break;
+	case COL_UUID:
+		data = udev_device_get_property_value(dev, "ID_FS_UUID_ENC");
+		break;
+	case COL_PARTUUID:
+		data = udev_device_get_property_value(dev, "ID_PART_ENTRY_UUID");
+		break;
+	case COL_PARTLABEL:
+		data = udev_device_get_property_value(dev, "ID_PART_ENTRY_NAME");
+		break;
+	default:
+		break;
+	}
+
+	if (data) {
+		res = xstrdup(data);
+		unhexmangle_string(res);
+	}
+
+	udev_device_unref(dev);
+	return res;
+}
+#endif /* HAVE_LIBUDEV */
+
+/* Returns LABEL or UUID */
+static const char *get_tag(struct libmnt_fs *fs, const char *tagname, int col
+#ifndef HAVE_LIBUDEV
+		__attribute__((__unused__))
+#endif
+		)
+{
+	const char *t, *v, *res = NULL;
 
 	if (!mnt_fs_get_tag(fs, &t, &v) && !strcmp(t, tagname))
 		res = v;
 	else {
-		res = mnt_fs_get_source(fs);
-		if (res)
-			res = mnt_resolve_spec(res, cache);
-		if (res)
-			res = mnt_cache_find_tag_value(cache, res, tagname);
+		const char *dev = mnt_fs_get_source(fs);
+
+		if (dev)
+			dev = mnt_resolve_spec(dev, cache);
+#ifdef HAVE_LIBUDEV
+		if (dev)
+			res = get_tag_from_udev(dev, col);
+#endif
+		if (!res)
+			res = mnt_cache_find_tag_value(cache, dev, tagname);
 	}
 
 	return res;
 }
 
+static const char *get_vfs_attr(struct libmnt_fs *fs, int sizetype)
+{
+	struct statvfs buf;
+	uint64_t vfs_attr = 0;
+	char *sizestr;
+
+	if (statvfs(mnt_fs_get_target(fs), &buf) != 0)
+		return NULL;
+
+	switch(sizetype) {
+	case COL_SIZE:
+		vfs_attr = buf.f_frsize * buf.f_blocks;
+		break;
+	case COL_AVAIL:
+		vfs_attr = buf.f_frsize * buf.f_bavail;
+		break;
+	case COL_USED:
+		vfs_attr = buf.f_frsize * (buf.f_blocks - buf.f_bfree);
+		break;
+	case COL_USEPERC:
+		if (buf.f_blocks == 0)
+			return "-";
+
+		xasprintf(&sizestr, "%.0f%%",
+				(double)(buf.f_blocks - buf.f_bfree) /
+				buf.f_blocks * 100);
+		return sizestr;
+	}
+
+	return vfs_attr == 0 ? "0" :
+		size_to_human_string(SIZE_SUFFIX_1LETTER, vfs_attr);
+}
+
 /* reads FS data from libmount
- * TODO: add function that will deallocate data allocated by get_data()
  */
 static const char *get_data(struct libmnt_fs *fs, int num)
 {
 	const char *str = NULL;
+	char *tmp;
+	int col_id = get_column_id(num);
 
-	switch(get_column_id(num)) {
+	switch (col_id) {
 	case COL_SOURCE:
 	{
 		const char *root = mnt_fs_get_root(fs);
@@ -297,10 +500,8 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 				str = mnt_resolve_spec(str, cache);
 		}
 		if (root && str && !(flags & FL_NOFSROOT) && strcmp(root, "/")) {
-			char *tmp;
-
-			if (asprintf(&tmp, "%s[%s]", str, root) > 0)
-				str = tmp;
+			xasprintf(&tmp, "%s[%s]", str, root);
+			str = tmp;
 		}
 		break;
 	}
@@ -319,28 +520,91 @@ static const char *get_data(struct libmnt_fs *fs, int num)
 	case COL_FS_OPTIONS:
 		str = mnt_fs_get_fs_options(fs);
 		break;
+	case COL_OPT_FIELDS:
+		str = mnt_fs_get_optional_fields(fs);
+		break;
 	case COL_UUID:
-		str = get_tag(fs, "UUID");
+		str = get_tag(fs, "UUID", col_id);
+		break;
+	case COL_PARTUUID:
+		str = get_tag(fs, "PARTUUID", col_id);
 		break;
 	case COL_LABEL:
-		str = get_tag(fs, "LABEL");
+		str = get_tag(fs, "LABEL", col_id);
 		break;
+	case COL_PARTLABEL:
+		str = get_tag(fs, "PARTLABEL", col_id);
+		break;
+
 	case COL_MAJMIN:
 	{
 		dev_t devno = mnt_fs_get_devno(fs);
-		if (devno) {
-			char *tmp;
-			int rc = 0;
-			if ((tt_flags & TT_FL_RAW) || (tt_flags & TT_FL_EXPORT))
-				rc = asprintf(&tmp, "%u:%u",
-					      major(devno), minor(devno));
-			else
-				rc = asprintf(&tmp, "%3u:%-3u",
-					      major(devno), minor(devno));
-			if (rc)
-				str = tmp;
-		}
+		if (!devno)
+			break;
+
+		if ((tt_flags & TT_FL_RAW) || (tt_flags & TT_FL_EXPORT))
+			xasprintf(&tmp, "%u:%u", major(devno), minor(devno));
+		else
+			xasprintf(&tmp, "%3u:%-3u", major(devno), minor(devno));
+		str = tmp;
+		break;
 	}
+	case COL_SIZE:
+	case COL_AVAIL:
+	case COL_USED:
+	case COL_USEPERC:
+		str = get_vfs_attr(fs, col_id);
+		break;
+	case COL_FSROOT:
+		str = mnt_fs_get_root(fs);
+		break;
+	case COL_TID:
+		if (mnt_fs_get_tid(fs)) {
+			xasprintf(&tmp, "%d", mnt_fs_get_tid(fs));
+			str = tmp;
+		}
+		break;
+	case COL_ID:
+		if (mnt_fs_get_id(fs)) {
+			xasprintf(&tmp, "%d", mnt_fs_get_id(fs));
+			str = tmp;
+		}
+		break;
+	case COL_PROPAGATION:
+		if (mnt_fs_is_kernel(fs)) {
+			unsigned long fl = 0;
+			char *n = NULL;
+
+			if (mnt_fs_get_propagation(fs, &fl) != 0)
+				break;
+
+			n = xstrdup((fl & MS_SHARED) ? "shared" : "private");
+
+			if (fl & MS_SLAVE) {
+				xasprintf(&tmp, "%s,slave", n);
+				free(n);
+				n = tmp;
+			}
+			if (fl & MS_UNBINDABLE) {
+				xasprintf(&tmp, "%s,unbindable", n);
+				free(n);
+				n = tmp;
+			}
+			str = n;
+		}
+		break;
+	case COL_FREQ:
+		if (!mnt_fs_is_kernel(fs)) {
+			xasprintf(&tmp, "%d", mnt_fs_get_freq(fs));
+			str = tmp;
+		}
+		break;
+	case COL_PASSNO:
+		if (!mnt_fs_is_kernel(fs)) {
+			xasprintf(&tmp, "%d", mnt_fs_get_passno(fs));
+			str = tmp;
+		}
+		break;
 	default:
 		break;
 	}
@@ -463,9 +727,12 @@ static int create_treenode(struct tt *tt, struct libmnt_table *tb,
 	if (!itr)
 		goto leave;
 
-	line = add_line(tt, fs, parent_line);
-	if (!line)
-		goto leave;
+	if ((flags & FL_SUBMOUNTS) || match_func(fs, NULL)) {
+		line = add_line(tt, fs, parent_line);
+		if (!line)
+			goto leave;
+	} else
+		line = parent_line;
 
 	/*
 	 * add all children to the output table
@@ -484,44 +751,108 @@ leave:
 static int parser_errcb(struct libmnt_table *tb __attribute__ ((__unused__)),
 			const char *filename, int line)
 {
-	warn(_("%s: parse error at line %d"), filename, line);
+	warnx(_("%s: parse error at line %d"), filename, line);
 	return 0;
 }
 
-/* calls libmount fstab/mtab/mountinfo parser */
-static struct libmnt_table *parse_tabfile(const char *path)
+static char **append_tabfile(char **files, int *nfiles, char *filename)
 {
-	int rc;
-	struct libmnt_table *tb = mnt_new_table();
+	files = xrealloc(files, sizeof(char *) * (*nfiles + 1));
+	files[(*nfiles)++] = filename;
+	return files;
+}
 
+static char **append_pid_tabfile(char **files, int *nfiles, pid_t pid)
+{
+	char *path = NULL;
+
+	xasprintf(&path, "/proc/%d/mountinfo", (int) pid);
+	return append_tabfile(files, nfiles, path);
+}
+
+/* calls libmount fstab/mtab/mountinfo parser */
+static struct libmnt_table *parse_tabfiles(char **files,
+					   int nfiles,
+					   int tabtype)
+{
+	struct libmnt_table *tb;
+	int rc = 0;
+
+	tb = mnt_new_table();
 	if (!tb) {
 		warn(_("failed to initialize libmount table"));
 		return NULL;
 	}
-
 	mnt_table_set_parser_errcb(tb, parser_errcb);
 
-	if (!strcmp(path, _PATH_MNTTAB))
-		rc = mnt_table_parse_fstab(tb, NULL);
-	else if (!strcmp(path, _PATH_MOUNTED))
-		rc = mnt_table_parse_mtab(tb, NULL);
-	else
-		rc = mnt_table_parse_file(tb, path);
+	do {
+		/* NULL means that libmount will use default paths */
+		const char *path = nfiles ? *files++ : NULL;
+
+		switch (tabtype) {
+		case TABTYPE_FSTAB:
+			rc = mnt_table_parse_fstab(tb, path);
+			break;
+		case TABTYPE_MTAB:
+			rc = mnt_table_parse_mtab(tb, path);
+			break;
+		case TABTYPE_KERNEL:
+			if (!path)
+				path = access(_PATH_PROC_MOUNTINFO, R_OK) == 0 ?
+					      _PATH_PROC_MOUNTINFO :
+					      _PATH_PROC_MOUNTS;
+
+			rc = mnt_table_parse_file(tb, path);
+			break;
+		}
+		if (rc) {
+			mnt_free_table(tb);
+			warn(_("can't read %s"), path);
+			return NULL;
+		}
+	} while (--nfiles > 0);
 
-	if (rc) {
-		mnt_free_table(tb);
-		warn(_("can't read %s"), path);
-		return NULL;
-	}
 	return tb;
 }
 
+/* checks if @tb contains parent->child relations */
+static int tab_is_tree(struct libmnt_table *tb)
+{
+	struct libmnt_fs *fs = NULL;
+	struct libmnt_iter *itr = NULL;
+	int rc = 0;
+
+	itr = mnt_new_iter(MNT_ITER_BACKWARD);
+	if (!itr)
+		return 0;
+
+	if (mnt_table_next_fs(tb, itr, &fs) == 0)
+		rc = mnt_fs_get_id(fs) > 0 && mnt_fs_get_parent_id(fs) > 0;
+
+	mnt_free_iter(itr);
+	return rc;
+}
+
+
 /* filter function for libmount (mnt_table_find_next_fs()) */
 static int match_func(struct libmnt_fs *fs,
 		      void *data __attribute__ ((__unused__)))
 {
 	int rc = flags & FL_INVERT ? 1 : 0;
 	const char *m;
+	void *md;
+
+	m = get_match(COL_FSTYPE);
+	if (m && !mnt_fs_match_fstype(fs, m))
+		return rc;
+
+	m = get_match(COL_OPTIONS);
+	if (m && !mnt_fs_match_options(fs, m))
+		return rc;
+
+	md = get_match_data(COL_MAJMIN);
+	if (md && mnt_fs_get_devno(fs) != *((dev_t *) md))
+		return rc;
 
 	m = get_match(COL_TARGET);
 	if (m && !mnt_fs_match_target(fs, m, cache))
@@ -531,13 +862,15 @@ static int match_func(struct libmnt_fs *fs,
 	if (m && !mnt_fs_match_source(fs, m, cache))
 		return rc;
 
-	m = get_match(COL_FSTYPE);
-	if (m && !mnt_fs_match_fstype(fs, m))
-		return rc;
+	if ((flags & FL_DF) && !(flags & FL_ALL)) {
+		const char *type = mnt_fs_get_fstype(fs);
 
-	m = get_match(COL_OPTIONS);
-	if (m && !mnt_fs_match_options(fs, m))
-		return rc;
+		if (type && strstr(type, "tmpfs"))	/* tmpfs is wanted */
+			return !rc;
+
+		if (mnt_fs_is_pseudofs(fs))
+			return rc;
+	}
 
 	return !rc;
 }
@@ -593,6 +926,10 @@ again:
 	return fs;
 }
 
+/*
+ * Filter out unwanted lines for --list output or top level lines for
+ * --submounts tree output.
+ */
 static int add_matching_lines(struct libmnt_table *tb,
 			      struct tt *tt, int direction)
 {
@@ -634,7 +971,7 @@ static int poll_match(struct libmnt_fs *fs)
 	    get_match(COL_SOURCE) && !get_match(COL_TARGET)) {
 		/*
 		 * findmnt --poll /foo
-		 * The '/foo' source as well as target.
+		 * The '/foo' maybe source as well as target.
 		 */
 		const char *str = get_match(COL_SOURCE);
 
@@ -653,7 +990,7 @@ static int poll_match(struct libmnt_fs *fs)
 static int poll_table(struct libmnt_table *tb, const char *tabfile,
 		  int timeout, struct tt *tt, int direction)
 {
-	FILE *f;
+	FILE *f = NULL;
 	int rc = -1;
 	struct libmnt_iter *itr = NULL;
 	struct libmnt_table *tb_new = NULL;
@@ -684,7 +1021,7 @@ static int poll_table(struct libmnt_table *tb, const char *tabfile,
 
 	f = fopen(tabfile, "r");
 	if (!f) {
-		warn(_("%s: open failed"), tabfile);
+		warn(_("cannot open %s"), tabfile);
 		goto done;
 	}
 
@@ -753,6 +1090,8 @@ done:
 	mnt_free_table(tb_new);
 	mnt_free_tabdiff(diff);
 	mnt_free_iter(itr);
+	if (f)
+		fclose(f);
 	return rc;
 }
 
@@ -760,8 +1099,8 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 {
 	int i;
 
+	fputs(USAGE_HEADER, out);
 	fprintf(out, _(
-	"\nUsage:\n"
 	" %1$s [options]\n"
 	" %1$s [options] <device> | <mountpoint>\n"
 	" %1$s [options] <device> <mountpoint>\n"
@@ -780,63 +1119,65 @@ static void __attribute__((__noreturn__)) usage(FILE *out)
 	" -w, --timeout <num>    upper limit in milliseconds that --poll will block\n\n"));
 
 	fprintf(out, _(
+	" -A, --all              disable all built-in filters, print all filesystems\n"
 	" -a, --ascii            use ASCII chars for tree formatting\n"
 	" -c, --canonicalize     canonicalize printed paths\n"
+	" -D, --df               imitate the output of df(1)\n"
 	" -d, --direction <word> direction of search, 'forward' or 'backward'\n"
-	" -e, --evaluate         convert tags (LABEL/UUID) to device names\n"
+	" -e, --evaluate         convert tags (LABEL,UUID,PARTUUID,PARTLABEL) \n"
+	"                          to device names\n"
+	" -F, --tab-file <path>  alternative file for --fstab, --mtab or --kernel options\n"
 	" -f, --first-only       print the first found filesystem only\n"));
 
 	fprintf(out, _(
-	" -h, --help             display this help text and exit\n"
 	" -i, --invert           invert the sense of matching\n"
 	" -l, --list             use list format output\n"
+	" -N, --task <tid>       use alternative namespace (/proc/<tid>/mountinfo file)\n"
 	" -n, --noheadings       don't print column headings\n"
 	" -u, --notruncate       don't truncate text in columns\n"));
-	fprintf(out, _(
-	" -O, --options <list>   limit the set of filesystems by mount options\n"
-	" -o, --output <list>    the output columns to be shown\n"
-	" -P, --pairs            use key=\"value\" output format\n"
-	" -r, --raw              use raw output format\n"
-	" -t, --types <list>     limit the set of filesystems by FS types\n"));
+fputs (_(" -O, --options <list>   limit the set of filesystems by mount options\n"), out);
+fputs (_(" -o, --output <list>    the output columns to be shown\n"), out);
+fputs (_(" -P, --pairs            use key=\"value\" output format\n"), out);
+fputs (_(" -r, --raw              use raw output format\n"), out);
+fputs (_(" -t, --types <list>     limit the set of filesystems by FS types\n"), out);
 	fprintf(out, _(
 	" -v, --nofsroot         don't print [/dir] for bind or btrfs mounts\n"
 	" -R, --submounts        print all submounts for the matching filesystems\n"
-	" -S, --source <string>  the device to mount (by name, LABEL= or UUID=)\n"
-	" -T, --target <string>  the mountpoint to use\n\n"));
+	" -S, --source <string>  the device to mount (by name, maj:min, \n"
+	"                          LABEL=, UUID=, PARTUUID=, PARTLABEL=)\n"
+	" -T, --target <string>  the mountpoint to use\n"));
 
+	fputs(USAGE_SEPARATOR, out);
+	fputs(USAGE_HELP, out);
+	fputs(USAGE_VERSION, out);
 
 	fprintf(out, _("\nAvailable columns:\n"));
 
 	for (i = 0; i < FINDMNT_NCOLUMNS; i++)
 		fprintf(out, " %11s  %s\n", infos[i].name, _(infos[i].help));
 
-
-	fprintf(out, _("\nFor more information see findmnt(1).\n"));
+	fprintf(out, USAGE_MAN_TAIL("findmnt(8)"));
 
 	exit(out == stderr ? EXIT_FAILURE : EXIT_SUCCESS);
 }
 
-static void __attribute__((__noreturn__))
-errx_mutually_exclusive(const char *opts)
-{
-	errx(EXIT_FAILURE, "%s %s", opts, _("options are mutually exclusive"));
-}
-
 int main(int argc, char *argv[])
 {
-	/* libmount */
 	struct libmnt_table *tb = NULL;
-	char *tabfile = NULL;
+	char **tabfiles = NULL;
 	int direction = MNT_ITER_FORWARD;
 	int i, c, rc = -1, timeout = -1;
+	int ntabfiles = 0, tabtype = 0;
+	char *outarg = NULL;
 
-	/* table.h */
 	struct tt *tt = NULL;
 
 	static const struct option longopts[] = {
+	    { "all",          0, 0, 'A' },
 	    { "ascii",        0, 0, 'a' },
 	    { "canonicalize", 0, 0, 'c' },
 	    { "direction",    1, 0, 'd' },
+	    { "df",           0, 0, 'D' },
 	    { "evaluate",     0, 0, 'e' },
 	    { "first-only",   0, 0, 'f' },
 	    { "fstab",        0, 0, 's' },
@@ -853,33 +1194,56 @@ int main(int argc, char *argv[])
 	    { "pairs",        0, 0, 'P' },
 	    { "raw",          0, 0, 'r' },
 	    { "types",        1, 0, 't' },
-	    { "fsroot",       0, 0, 'v' },
+	    { "nofsroot",     0, 0, 'v' },
 	    { "submounts",    0, 0, 'R' },
 	    { "source",       1, 0, 'S' },
+	    { "tab-file",     1, 0, 'F' },
+	    { "task",         1, 0, 'N' },
 	    { "target",       1, 0, 'T' },
 	    { "timeout",      1, 0, 'w' },
+	    { "version",      0, 0, 'V' },
 
 	    { NULL,           0, 0, 0 }
 	};
 
+	static const ul_excl_t excl[] = {	/* rows and cols in in ASCII order */
+		{ 'N','k','m','s' },		/* task,kernel,mtab,fstab */
+		{ 'P','l','r' },		/* pairs,list,raw */
+		{ 'm','p','s' },		/* mtab,poll,fstab */
+		{ 0 }
+	};
+	int excl_st[ARRAY_SIZE(excl)] = UL_EXCL_STATUS_INIT;
+
 	assert(ARRAY_SIZE(columns) == FINDMNT_NCOLUMNS);
 
 	setlocale(LC_ALL, "");
 	bindtextdomain(PACKAGE, LOCALEDIR);
 	textdomain(PACKAGE);
+	atexit(close_stdout);
 
 	/* default output format */
 	tt_flags |= TT_FL_TREE;
 
 	while ((c = getopt_long(argc, argv,
-				"acd:ehifo:O:p::Pklmnrst:uvRS:T:w:", longopts, NULL)) != -1) {
+				"AacDd:ehifF:o:O:p::PklmnN:rst:uvRS:T:w:V",
+				longopts, NULL)) != -1) {
+
+		err_exclusive_options(c, longopts, excl, excl_st);
+
 		switch(c) {
+		case 'A':
+			flags |= FL_ALL;
+			break;
 		case 'a':
 			tt_flags |= TT_FL_ASCII;
 			break;
 		case 'c':
 			flags |= FL_CANONICALIZE;
 			break;
+		case 'D':
+			tt_flags &= ~TT_FL_TREE;
+			flags |= FL_DF;
+			break;
 		case 'd':
 			if (!strcmp(optarg, "forward"))
 				direction = MNT_ITER_FORWARD;
@@ -901,15 +1265,14 @@ int main(int argc, char *argv[])
 		case 'f':
 			flags |= FL_FIRSTONLY;
 			break;
+		case 'F':
+			tabfiles = append_tabfile(tabfiles, &ntabfiles, optarg);
+			break;
 		case 'u':
 			disable_columns_truncate();
 			break;
 		case 'o':
-			ncolumns = string_to_idarray(optarg,
-						columns, ARRAY_SIZE(columns),
-						column_name_to_id);
-			if (ncolumns < 0)
-				exit(EXIT_FAILURE);
+			outarg = optarg;
 			break;
 		case 'O':
 			set_match(COL_OPTIONS, optarg);
@@ -930,21 +1293,15 @@ int main(int argc, char *argv[])
 			tt_flags &= ~TT_FL_TREE;
 			break;
 		case 'm':		/* mtab */
-			if (tabfile)
-				errx_mutually_exclusive("--{fstab,mtab,kernel}");
-			tabfile = _PATH_MOUNTED;
+			tabtype = TABTYPE_MTAB;
 			tt_flags &= ~TT_FL_TREE;
 			break;
 		case 's':		/* fstab */
-			if (tabfile)
-				errx_mutually_exclusive("--{fstab,mtab,kernel}");
-			tabfile = _PATH_MNTTAB;
+			tabtype = TABTYPE_FSTAB;
 			tt_flags &= ~TT_FL_TREE;
 			break;
 		case 'k':		/* kernel (mountinfo) */
-			if (tabfile)
-				 errx_mutually_exclusive("--{fstab,mtab,kernel}");
-			tabfile = _PATH_PROC_MOUNTINFO;
+			tabtype = TABTYPE_KERNEL;
 			break;
 		case 't':
 			set_match(COL_FSTYPE, optarg);
@@ -954,14 +1311,17 @@ int main(int argc, char *argv[])
 			tt_flags |= TT_FL_RAW;		/* enable raw */
 			break;
 		case 'l':
-			if ((tt_flags & TT_FL_RAW) && (tt_flags & TT_FL_EXPORT))
-				errx_mutually_exclusive("--{raw,list,pairs}");
-
 			tt_flags &= ~TT_FL_TREE; /* disable the default */
 			break;
 		case 'n':
 			tt_flags |= TT_FL_NOHEADINGS;
 			break;
+		case 'N':
+			tabtype = TABTYPE_KERNEL;
+			tabfiles = append_pid_tabfile(tabfiles, &ntabfiles,
+					strtou32_or_err(optarg,
+						_("invalid TID argument")));
+			break;
 		case 'v':
 			flags |= FL_NOFSROOT;
 			break;
@@ -969,7 +1329,7 @@ int main(int argc, char *argv[])
 			flags |= FL_SUBMOUNTS;
 			break;
 		case 'S':
-			set_match(COL_SOURCE, optarg);
+			set_source_match(optarg);
 			flags |= FL_NOSWAPMATCH;
 			break;
 		case 'T':
@@ -977,15 +1337,27 @@ int main(int argc, char *argv[])
 			flags |= FL_NOSWAPMATCH;
 			break;
 		case 'w':
-			timeout = strtol_or_err(optarg,
-					_("failed to parse timeout"));
+			timeout = strtos32_or_err(optarg, _("invalid timeout argument"));
 			break;
+		case 'V':
+			printf(UTIL_LINUX_VERSION);
+			return EXIT_SUCCESS;
 		default:
 			usage(stderr);
 			break;
 		}
 	}
 
+	if (!ncolumns && (flags & FL_DF)) {
+		columns[ncolumns++] = COL_SOURCE;
+		columns[ncolumns++] = COL_FSTYPE;
+		columns[ncolumns++] = COL_SIZE;
+		columns[ncolumns++] = COL_USED;
+		columns[ncolumns++] = COL_AVAIL;
+		columns[ncolumns++] = COL_USEPERC;
+		columns[ncolumns++] = COL_TARGET;
+	}
+
 	/* default columns */
 	if (!ncolumns) {
 		if (flags & FL_POLL)
@@ -997,14 +1369,15 @@ int main(int argc, char *argv[])
 		columns[ncolumns++] = COL_OPTIONS;
 	}
 
-	if (!tabfile) {
-		tabfile = _PATH_PROC_MOUNTINFO;
+	if (outarg && string_add_to_idarray(outarg, columns, ARRAY_SIZE(columns),
+					 &ncolumns, column_name_to_id) < 0)
+		return EXIT_FAILURE;
 
-		if (access(tabfile, R_OK)) {		/* old kernel? */
-			tabfile = _PATH_PROC_MOUNTS;
-			tt_flags &= ~TT_FL_TREE;
-		}
-	}
+	if (!tabtype)
+		tabtype = TABTYPE_KERNEL;
+
+	if ((flags & FL_POLL) && ntabfiles > 1)
+		errx(EXIT_FAILURE, _("--poll accepts only one file, but more specified by --tab-file"));
 
 	if (optind < argc && (get_match(COL_SOURCE) || get_match(COL_TARGET)))
 		errx(EXIT_FAILURE, _(
@@ -1012,7 +1385,7 @@ int main(int argc, char *argv[])
 			"with command line element that is not an option"));
 
 	if (optind < argc)
-		set_match(COL_SOURCE, argv[optind++]);	/* dev/tag/mountpoint */
+		set_source_match(argv[optind++]);	/* dev/tag/mountpoint/maj:min */
 	if (optind < argc)
 		set_match(COL_TARGET, argv[optind++]);	/* mountpoint */
 
@@ -1020,8 +1393,10 @@ int main(int argc, char *argv[])
 		/* don't care about submounts if list all mounts */
 		flags &= ~FL_SUBMOUNTS;
 
-	if (!(flags & FL_SUBMOUNTS) &&
-	    (!is_listall_mode() || (flags & FL_FIRSTONLY)))
+	if (!(flags & FL_SUBMOUNTS) && ((flags & FL_FIRSTONLY)
+	    || get_match(COL_TARGET)
+	    || get_match(COL_SOURCE)
+	    || get_match(COL_MAJMIN)))
 		tt_flags &= ~TT_FL_TREE;
 
 	if (!(flags & FL_NOSWAPMATCH) &&
@@ -1032,7 +1407,8 @@ int main(int argc, char *argv[])
 		 */
 		const char *x = get_match(COL_SOURCE);
 
-		if (!strncmp(x, "LABEL=", 6) || !strncmp(x, "UUID=", 5))
+		if (!strncmp(x, "LABEL=", 6) || !strncmp(x, "UUID=", 5) ||
+		    !strncmp(x, "PARTLABEL=", 10) || !strncmp(x, "PARTUUID=", 9))
 			flags |= FL_NOSWAPMATCH;
 	}
 
@@ -1041,10 +1417,13 @@ int main(int argc, char *argv[])
 	 */
 	mnt_init_debug(0);
 
-	tb = parse_tabfile(tabfile);
+	tb = parse_tabfiles(tabfiles, ntabfiles, tabtype);
 	if (!tb)
 		goto leave;
 
+	if ((tt_flags & TT_FL_TREE) && (ntabfiles > 1 || !tab_is_tree(tb)))
+		tt_flags &= ~TT_FL_TREE;
+
 	cache = mnt_new_cache();
 	if (!cache) {
 		warn(_("failed to initialize libmount cache"));
@@ -1052,6 +1431,7 @@ int main(int argc, char *argv[])
 	}
 	mnt_table_set_cache(tb, cache);
 
+
 	/*
 	 * initialize output formatting (tt.h)
 	 */
@@ -1083,17 +1463,31 @@ int main(int argc, char *argv[])
 	/*
 	 * Fill in data to the output table
 	 */
-	if (flags & FL_POLL)
-		/* poll mode */
-		rc = poll_table(tb, tabfile, timeout, tt, direction);
+	if (flags & FL_POLL) {
+		/* poll mode (accept the first tabfile only) */
+		rc = poll_table(tb, tabfiles ? *tabfiles : _PATH_PROC_MOUNTINFO, timeout, tt, direction);
 
-	else if ((tt_flags & TT_FL_TREE) && is_listall_mode())
+	} else if ((tt_flags & TT_FL_TREE) && !(flags & FL_SUBMOUNTS)) {
 		/* whole tree */
 		rc = create_treenode(tt, tb, NULL, NULL);
-	else
+	} else {
 		/* whole lits of sub-tree */
 		rc = add_matching_lines(tb, tt, direction);
 
+		if (rc != 0
+		    && tabtype == TABTYPE_KERNEL
+		    && (flags & FL_NOSWAPMATCH)
+		    && get_match(COL_TARGET)) {
+			/*
+			 * Found nothing, maybe the --target is regular file,
+			 * try it again with extra functionality for target
+			 * match
+			 */
+			enable_extra_target_match();
+			rc = add_matching_lines(tb, tt, direction);
+		}
+	}
+
 	/*
 	 * Print the output table for non-poll modes
 	 */
@@ -1104,6 +1498,9 @@ leave:
 
 	mnt_free_table(tb);
 	mnt_free_cache(cache);
-
+	free(tabfiles);
+#ifdef HAVE_LIBUDEV
+	udev_unref(udev);
+#endif
 	return rc ? EXIT_FAILURE : EXIT_SUCCESS;
 }
